# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import os
import pytest


import swh.loader.package
from swh.loader.package.utils import download, api_info, release_name, artifact_identity


def test_version_generation():
    assert (
        swh.loader.package.__version__ != "devel"
    ), "Make sure swh.loader.core is installed (e.g. pip install -e .)"


@pytest.mark.fs
def test_download_fail_to_download(tmp_path, requests_mock):
    url = "https://pypi.org/pypi/arrow/json"
    status_code = 404
    requests_mock.get(url, status_code=status_code)

    with pytest.raises(ValueError) as e:
        download(url, tmp_path)

    assert e.value.args[0] == "Fail to query '%s'. Reason: %s" % (url, status_code)


@pytest.mark.fs
def test_download_ok(tmp_path, requests_mock):
    """Download without issue should provide filename and hashes"""
    filename = "requests-0.0.1.tar.gz"
    url = "https://pypi.org/pypi/requests/%s" % filename
    data = "this is something"
    requests_mock.get(url, text=data, headers={"content-length": str(len(data))})

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path))

    actual_filename = os.path.basename(actual_filepath)
    assert actual_filename == filename
    assert actual_hashes["length"] == len(data)
    assert (
        actual_hashes["checksums"]["sha1"] == "fdd1ce606a904b08c816ba84f3125f2af44d92b2"
    )  # noqa
    assert (
        actual_hashes["checksums"]["sha256"]
        == "1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5"
    )


@pytest.mark.fs
def test_download_ok_no_header(tmp_path, requests_mock):
    """Download without issue should provide filename and hashes"""
    filename = "requests-0.0.1.tar.gz"
    url = "https://pypi.org/pypi/requests/%s" % filename
    data = "this is something"
    requests_mock.get(url, text=data)  # no header information

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path))

    actual_filename = os.path.basename(actual_filepath)
    assert actual_filename == filename
    assert actual_hashes["length"] == len(data)
    assert (
        actual_hashes["checksums"]["sha1"] == "fdd1ce606a904b08c816ba84f3125f2af44d92b2"
    )  # noqa
    assert (
        actual_hashes["checksums"]["sha256"]
        == "1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5"
    )


@pytest.mark.fs
def test_download_ok_with_hashes(tmp_path, requests_mock):
    """Download without issue should provide filename and hashes"""
    filename = "requests-0.0.1.tar.gz"
    url = "https://pypi.org/pypi/requests/%s" % filename
    data = "this is something"
    requests_mock.get(url, text=data, headers={"content-length": str(len(data))})

    # good hashes for such file
    good = {
        "sha1": "fdd1ce606a904b08c816ba84f3125f2af44d92b2",
        "sha256": "1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5",  # noqa
    }

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path), hashes=good)

    actual_filename = os.path.basename(actual_filepath)
    assert actual_filename == filename
    assert actual_hashes["length"] == len(data)
    assert actual_hashes["checksums"]["sha1"] == good["sha1"]
    assert actual_hashes["checksums"]["sha256"] == good["sha256"]


@pytest.mark.fs
def test_download_fail_hashes_mismatch(tmp_path, requests_mock):
    """Mismatch hash after download should raise

    """
    filename = "requests-0.0.1.tar.gz"
    url = "https://pypi.org/pypi/requests/%s" % filename
    data = "this is something"
    requests_mock.get(url, text=data, headers={"content-length": str(len(data))})

    # good hashes for such file
    good = {
        "sha1": "fdd1ce606a904b08c816ba84f3125f2af44d92b2",
        "sha256": "1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5",  # noqa
    }

    for hash_algo in good.keys():
        wrong_hash = good[hash_algo].replace("1", "0")
        expected_hashes = good.copy()
        expected_hashes[hash_algo] = wrong_hash  # set the wrong hash

        expected_msg = "Failure when fetching %s. " "Checksum mismatched: %s != %s" % (
            url,
            wrong_hash,
            good[hash_algo],
        )

        with pytest.raises(ValueError, match=expected_msg):
            download(url, dest=str(tmp_path), hashes=expected_hashes)


def test_api_info_failure(requests_mock):
    """Failure to fetch info/release information should raise"""
    url = "https://pypi.org/pypi/requests/json"
    status_code = 400
    requests_mock.get(url, status_code=status_code)

    with pytest.raises(ValueError) as e0:
        api_info(url)

    assert e0.value.args[0] == "Fail to query '%s'. Reason: %s" % (url, status_code)


def test_api_info(requests_mock):
    """Fetching json info from pypi project should be ok"""
    url = "https://pypi.org/pypi/requests/json"
    requests_mock.get(url, text='{"version": "0.0.1"}')
    actual_info = api_info(url)
    assert actual_info == {
        "version": "0.0.1",
    }


def test_release_name():
    for version, filename, expected_release in [
        ("0.0.1", None, "releases/0.0.1"),
        ("0.0.2", "something", "releases/0.0.2/something"),
    ]:
        assert release_name(version, filename) == expected_release


def test_artifact_identity():
    """Compute primary key should return the right identity

    """
    data = {
        "a": 1,
        "b": 2,
        "length": 221837,
        "filename": "8sync-0.1.0.tar.gz",
        "version": "0.1.0",
    }

    for id_keys, expected_id in [
        (["a", "b"], [1, 2]),
        ([], []),
        (["a", "key-that-does-not-exist"], [1, None]),
    ]:
        actual_id = artifact_identity(data, id_keys=id_keys)
        assert actual_id == expected_id



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from typing import Any, Dict


@pytest.fixture
def swh_loader_config() -> Dict[str, Any]:
    return {
        "storage": {"cls": "pipeline", "steps": [{"cls": "memory",},],},
        "deposit": {
            "url": "https://deposit.softwareheritage.org/1/private",
            "auth": {"username": "user", "password": "pass",},
        },
    }



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import logging
import pytest
import random

from os import path

from swh.loader.package.debian.loader import (
    DebianLoader,
    download_package,
    dsc_information,
    uid_to_person,
    prepare_person,
    get_package_metadata,
    extract_package,
)
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
)

from swh.loader.package.debian.loader import resolve_revision_from

from swh.model.hashutil import hash_to_bytes
from swh.model.model import Person


logger = logging.getLogger(__name__)


PACKAGE_FILES = {
    "name": "cicero",
    "version": "0.7.2-3",
    "files": {
        "cicero_0.7.2-3.diff.gz": {
            "md5sum": "a93661b6a48db48d59ba7d26796fc9ce",
            "name": "cicero_0.7.2-3.diff.gz",
            "sha256": "f039c9642fe15c75bed5254315e2a29f9f2700da0e29d9b0729b3ffc46c8971c",  # noqa
            "size": 3964,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-3.diff.gz",  # noqa
        },
        "cicero_0.7.2-3.dsc": {
            "md5sum": "d5dac83eb9cfc9bb52a15eb618b4670a",
            "name": "cicero_0.7.2-3.dsc",
            "sha256": "35b7f1048010c67adfd8d70e4961aefd8800eb9a83a4d1cc68088da0009d9a03",  # noqa
            "size": 1864,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-3.dsc",  # noqa
        },  # noqa
        "cicero_0.7.2.orig.tar.gz": {
            "md5sum": "4353dede07c5728319ba7f5595a7230a",
            "name": "cicero_0.7.2.orig.tar.gz",
            "sha256": "63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786",  # noqa
            "size": 96527,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2.orig.tar.gz",  # noqa
        },
    },
}

PACKAGE_FILES2 = {
    "name": "cicero",
    "version": "0.7.2-4",
    "files": {
        "cicero_0.7.2-4.diff.gz": {
            "md5sum": "1e7e6fc4a59d57c98082a3af78145734",
            "name": "cicero_0.7.2-4.diff.gz",
            "sha256": "2e6fa296ee7005473ff58d0971f4fd325617b445671480e9f2cfb738d5dbcd01",  # noqa
            "size": 4038,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-4.diff.gz",  # noqa
        },
        "cicero_0.7.2-4.dsc": {
            "md5sum": "1a6c8855a73b4282bb31d15518f18cde",
            "name": "cicero_0.7.2-4.dsc",
            "sha256": "913ee52f7093913420de5cbe95d63cfa817f1a1daf997961149501894e754f8b",  # noqa
            "size": 1881,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-4.dsc",  # noqa
        },  # noqa
        "cicero_0.7.2.orig.tar.gz": {
            "md5sum": "4353dede07c5728319ba7f5595a7230a",
            "name": "cicero_0.7.2.orig.tar.gz",
            "sha256": "63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786",  # noqa
            "size": 96527,
            "uri": "http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2.orig.tar.gz",  # noqa
        },
    },
}


PACKAGE_PER_VERSION = {
    "stretch/contrib/0.7.2-3": PACKAGE_FILES,
}


PACKAGES_PER_VERSION = {
    "stretch/contrib/0.7.2-3": PACKAGE_FILES,
    "buster/contrib/0.7.2-4": PACKAGE_FILES2,
}


def test_debian_first_visit(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    url = "deb://Debian/packages/cicero"
    loader = DebianLoader(
        url=url, date="2019-10-12T05:58:09.165557+00:00", packages=PACKAGE_PER_VERSION,
    )

    actual_load_status = loader.load()
    expected_snapshot_id = "3b6b66e6ee4e7d903a379a882684a2a50480c0b4"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="deb")

    stats = get_stats(loader.storage)
    assert {
        "content": 42,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,  # all artifacts under 1 revision
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": {
            b"releases/stretch/contrib/0.7.2-3": {
                "target_type": "revision",
                "target": hash_to_bytes("2807f5b3f84368b4889a9ae827fe85854ffecf07"),
            }
        },
    }  # different than the previous loader as no release is done

    check_snapshot(expected_snapshot, loader.storage)


def test_debian_first_visit_then_another_visit(swh_config, requests_mock_datadir):
    """With no prior visit, load a debian project ends up with 1 snapshot

    """
    url = "deb://Debian/packages/cicero"
    loader = DebianLoader(
        url=url, date="2019-10-12T05:58:09.165557+00:00", packages=PACKAGE_PER_VERSION
    )

    actual_load_status = loader.load()

    expected_snapshot_id = "3b6b66e6ee4e7d903a379a882684a2a50480c0b4"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="deb")

    stats = get_stats(loader.storage)
    assert {
        "content": 42,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,  # all artifacts under 1 revision
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": {
            b"releases/stretch/contrib/0.7.2-3": {
                "target_type": "revision",
                "target": hash_to_bytes("2807f5b3f84368b4889a9ae827fe85854ffecf07"),
            }
        },
    }  # different than the previous loader as no release is done

    check_snapshot(expected_snapshot, loader.storage)

    # No change in between load
    actual_load_status2 = loader.load()
    assert actual_load_status2["status"] == "uneventful"
    assert_last_visit_matches(loader.storage, url, status="full", type="deb")

    stats2 = get_stats(loader.storage)
    assert {
        "content": 42 + 0,
        "directory": 2 + 0,
        "origin": 1,
        "origin_visit": 1 + 1,  # a new visit occurred
        "person": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,  # same snapshot across 2 visits
    } == stats2

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url.startswith("http://deb.debian.org")
    ]
    # visited each package artifact twice across 2 visits
    assert len(urls) == len(set(urls))


def test_uid_to_person():
    uid = "Someone Name <someone@orga.org>"
    actual_person = uid_to_person(uid)

    assert actual_person == {
        "name": "Someone Name",
        "email": "someone@orga.org",
        "fullname": uid,
    }


def test_prepare_person():
    actual_author = prepare_person(
        {
            "name": "Someone Name",
            "email": "someone@orga.org",
            "fullname": "Someone Name <someone@orga.org>",
        }
    )

    assert actual_author == Person(
        name=b"Someone Name",
        email=b"someone@orga.org",
        fullname=b"Someone Name <someone@orga.org>",
    )


def test_download_package(datadir, tmpdir, requests_mock_datadir):
    tmpdir = str(tmpdir)  # py3.5 work around (LocalPath issue)
    all_hashes = download_package(PACKAGE_FILES, tmpdir)
    assert all_hashes == {
        "cicero_0.7.2-3.diff.gz": {
            "checksums": {
                "sha1": "0815282053f21601b0ec4adf7a8fe47eace3c0bc",
                "sha256": "f039c9642fe15c75bed5254315e2a29f9f2700da0e29d9b0729b3ffc46c8971c",  # noqa
            },
            "filename": "cicero_0.7.2-3.diff.gz",
            "length": 3964,
        },
        "cicero_0.7.2-3.dsc": {
            "checksums": {
                "sha1": "abbec4e8efbbc80278236e1dd136831eac08accd",
                "sha256": "35b7f1048010c67adfd8d70e4961aefd8800eb9a83a4d1cc68088da0009d9a03",  # noqa
            },
            "filename": "cicero_0.7.2-3.dsc",
            "length": 1864,
        },
        "cicero_0.7.2.orig.tar.gz": {
            "checksums": {
                "sha1": "a286efd63fe2c9c9f7bb30255c3d6fcdcf390b43",
                "sha256": "63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786",  # noqa
            },
            "filename": "cicero_0.7.2.orig.tar.gz",
            "length": 96527,
        },
    }


def test_dsc_information_ok():
    fname = "cicero_0.7.2-3.dsc"
    dsc_url, dsc_name = dsc_information(PACKAGE_FILES)

    assert dsc_url == PACKAGE_FILES["files"][fname]["uri"]
    assert dsc_name == PACKAGE_FILES["files"][fname]["name"]


def test_dsc_information_not_found():
    fname = "cicero_0.7.2-3.dsc"
    package_files = copy.deepcopy(PACKAGE_FILES)
    package_files["files"].pop(fname)

    dsc_url, dsc_name = dsc_information(package_files)

    assert dsc_url is None
    assert dsc_name is None


def test_dsc_information_too_many_dsc_entries():
    # craft an extra dsc file
    fname = "cicero_0.7.2-3.dsc"
    package_files = copy.deepcopy(PACKAGE_FILES)
    data = package_files["files"][fname]
    fname2 = fname.replace("cicero", "ciceroo")
    package_files["files"][fname2] = data

    with pytest.raises(
        ValueError,
        match="Package %s_%s references several dsc"
        % (package_files["name"], package_files["version"]),
    ):
        dsc_information(package_files)


def test_get_package_metadata(requests_mock_datadir, datadir, tmp_path):
    tmp_path = str(tmp_path)  # py3.5 compat.
    package = PACKAGE_FILES

    logger.debug("package: %s", package)

    # download the packages
    all_hashes = download_package(package, tmp_path)

    # Retrieve information from package
    _, dsc_name = dsc_information(package)

    dl_artifacts = [(tmp_path, hashes) for hashes in all_hashes.values()]

    # Extract information from package
    extracted_path = extract_package(dl_artifacts, tmp_path)

    # Retrieve information on package
    dsc_path = path.join(path.dirname(extracted_path), dsc_name)
    actual_package_info = get_package_metadata(package, dsc_path, extracted_path)

    logger.debug("actual_package_info: %s", actual_package_info)

    assert actual_package_info == {
        "changelog": {
            "date": "2014-10-19T16:52:35+02:00",
            "history": [
                ("cicero", "0.7.2-2"),
                ("cicero", "0.7.2-1"),
                ("cicero", "0.7-1"),
            ],
            "person": {
                "email": "sthibault@debian.org",
                "fullname": "Samuel Thibault <sthibault@debian.org>",
                "name": "Samuel Thibault",
            },
        },
        "maintainers": [
            {
                "email": "debian-accessibility@lists.debian.org",
                "fullname": "Debian Accessibility Team "
                "<debian-accessibility@lists.debian.org>",
                "name": "Debian Accessibility Team",
            },
            {
                "email": "sthibault@debian.org",
                "fullname": "Samuel Thibault <sthibault@debian.org>",
                "name": "Samuel Thibault",
            },
        ],
        "name": "cicero",
        "version": "0.7.2-3",
    }


def test_debian_multiple_packages(swh_config, requests_mock_datadir):
    url = "deb://Debian/packages/cicero"
    loader = DebianLoader(
        url=url, date="2019-10-12T05:58:09.165557+00:00", packages=PACKAGES_PER_VERSION
    )

    actual_load_status = loader.load()
    expected_snapshot_id = "defc19021187f3727293121fcf6c5c82cb923604"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="deb")

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": {
            b"releases/stretch/contrib/0.7.2-3": {
                "target_type": "revision",
                "target": hash_to_bytes("2807f5b3f84368b4889a9ae827fe85854ffecf07"),
            },
            b"releases/buster/contrib/0.7.2-4": {
                "target_type": "revision",
                "target": hash_to_bytes("8224139c274c984147ef4b09aa0e462c55a10bd3"),
            },
        },
    }

    check_snapshot(expected_snapshot, loader.storage)


def test_resolve_revision_from_edge_cases():
    """Solving revision with empty data will result in unknown revision

    """
    for package_artifacts in [{}, PACKAGE_FILES]:
        actual_revision = resolve_revision_from(package_artifacts, {})
        assert actual_revision is None

    for known_artifacts in [{}, PACKAGE_FILES]:
        actual_revision = resolve_revision_from({}, known_artifacts)
        assert actual_revision is None

    known_package_artifacts = {
        b"(\x07\xf5\xb3\xf8Ch\xb4\x88\x9a\x9a\xe8'\xfe\x85\x85O\xfe\xcf\x07": {
            "extrinsic": {
                # empty
            },
            # ... removed the unnecessary intermediary data
        }
    }
    assert not resolve_revision_from(known_package_artifacts, PACKAGE_FILES)


def test_resolve_revision_from_edge_cases_hit_and_miss():
    """Solving revision with inconsistent data will result in unknown revision

    """
    artifact_metadata = PACKAGE_FILES2
    expected_revision_id = (
        b"(\x08\xf5\xb3\xf8Ch\xb4\x88\x9a\x9a\xe8'\xff\x85\x85O\xfe\xcf\x07"  # noqa
    )
    known_package_artifacts = {
        expected_revision_id: {
            "extrinsic": {"raw": PACKAGE_FILES,},
            # ... removed the unnecessary intermediary data
        }
    }

    actual_revision = resolve_revision_from(known_package_artifacts, artifact_metadata)

    assert actual_revision is None


def test_resolve_revision_from():
    """Solving revision with consistent data will solve the revision

    """
    artifact_metadata = PACKAGE_FILES
    expected_revision_id = (
        b"(\x07\xf5\xb3\xf8Ch\xb4\x88\x9a\x9a\xe8'\xfe\x85\x85O\xfe\xcf\x07"  # noqa
    )

    files = artifact_metadata["files"]
    # shuffling dict's keys
    keys = list(files.keys())
    random.shuffle(keys)
    package_files = {"files": {k: files[k] for k in keys}}

    known_package_artifacts = {
        expected_revision_id: {
            "extrinsic": {"raw": package_files,},
            # ... removed the unnecessary intermediary data
        }
    }

    actual_revision = resolve_revision_from(known_package_artifacts, artifact_metadata)

    assert actual_revision == expected_revision_id



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from os import path

import pytest

from unittest.mock import patch

from swh.core.tarball import uncompress
from swh.core.pytest_plugin import requests_mock_datadir_factory
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Person

from swh.loader.package.pypi.loader import (
    PyPILoader,
    pypi_api_url,
    author,
    extract_intrinsic_metadata,
    artifact_to_revision_id,
)
from swh.loader.package.tests.common import check_metadata_paths
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
)


def test_author_basic():
    data = {
        "author": "i-am-groot",
        "author_email": "iam@groot.org",
    }
    actual_author = author(data)

    expected_author = Person(
        fullname=b"i-am-groot <iam@groot.org>",
        name=b"i-am-groot",
        email=b"iam@groot.org",
    )

    assert actual_author == expected_author


def test_author_empty_email():
    data = {
        "author": "i-am-groot",
        "author_email": "",
    }
    actual_author = author(data)

    expected_author = Person(fullname=b"i-am-groot", name=b"i-am-groot", email=b"",)

    assert actual_author == expected_author


def test_author_empty_name():
    data = {
        "author": "",
        "author_email": "iam@groot.org",
    }
    actual_author = author(data)

    expected_author = Person(
        fullname=b" <iam@groot.org>", name=b"", email=b"iam@groot.org",
    )

    assert actual_author == expected_author


def test_author_malformed():
    data = {
        "author": "['pierre', 'paul', 'jacques']",
        "author_email": None,
    }

    actual_author = author(data)

    expected_author = Person(
        fullname=b"['pierre', 'paul', 'jacques']",
        name=b"['pierre', 'paul', 'jacques']",
        email=None,
    )

    assert actual_author == expected_author


def test_author_malformed_2():
    data = {
        "author": "[marie, jeanne]",
        "author_email": "[marie@some, jeanne@thing]",
    }

    actual_author = author(data)

    expected_author = Person(
        fullname=b"[marie, jeanne] <[marie@some, jeanne@thing]>",
        name=b"[marie, jeanne]",
        email=b"[marie@some, jeanne@thing]",
    )

    assert actual_author == expected_author


def test_author_malformed_3():
    data = {
        "author": "[marie, jeanne, pierre]",
        "author_email": "[marie@somewhere.org, jeanne@somewhere.org]",
    }

    actual_author = author(data)

    expected_author = Person(
        fullname=(
            b"[marie, jeanne, pierre] " b"<[marie@somewhere.org, jeanne@somewhere.org]>"
        ),
        name=b"[marie, jeanne, pierre]",
        email=b"[marie@somewhere.org, jeanne@somewhere.org]",
    )

    actual_author == expected_author


# configuration error #


def test_badly_configured_loader_raise(monkeypatch):
    """Badly configured loader should raise"""
    monkeypatch.delenv("SWH_CONFIG_FILENAME", raising=False)
    with pytest.raises(ValueError) as e:
        PyPILoader(url="some-url")

    assert "Misconfiguration" in e.value.args[0]


def test_pypi_api_url():
    """Compute pypi api url from the pypi project url should be ok"""
    url = pypi_api_url("https://pypi.org/project/requests")
    assert url == "https://pypi.org/pypi/requests/json"


def test_pypi_api_url_with_slash():
    """Compute pypi api url from the pypi project url should be ok"""
    url = pypi_api_url("https://pypi.org/project/requests/")
    assert url == "https://pypi.org/pypi/requests/json"


@pytest.mark.fs
def test_extract_intrinsic_metadata(tmp_path, datadir):
    """Parsing existing archive's PKG-INFO should yield results"""
    uncompressed_archive_path = str(tmp_path)
    archive_path = path.join(
        datadir, "https_files.pythonhosted.org", "0805nexter-1.1.0.zip"
    )
    uncompress(archive_path, dest=uncompressed_archive_path)

    actual_metadata = extract_intrinsic_metadata(uncompressed_archive_path)
    expected_metadata = {
        "metadata_version": "1.0",
        "name": "0805nexter",
        "version": "1.1.0",
        "summary": "a simple printer of nested lest",
        "home_page": "http://www.hp.com",
        "author": "hgtkpython",
        "author_email": "2868989685@qq.com",
        "platforms": ["UNKNOWN"],
    }

    assert actual_metadata == expected_metadata


@pytest.mark.fs
def test_extract_intrinsic_metadata_failures(tmp_path):
    """Parsing inexistent path/archive/PKG-INFO yield None"""
    tmp_path = str(tmp_path)  # py3.5 work around (PosixPath issue)
    # inexistent first level path
    assert extract_intrinsic_metadata("/something-inexistent") == {}
    # inexistent second level path (as expected by pypi archives)
    assert extract_intrinsic_metadata(tmp_path) == {}
    # inexistent PKG-INFO within second level path
    existing_path_no_pkginfo = path.join(tmp_path, "something")
    os.mkdir(existing_path_no_pkginfo)
    assert extract_intrinsic_metadata(tmp_path) == {}


# LOADER SCENARIO #

# "edge" cases (for the same origin) #


# no release artifact:
# {visit full, status: uneventful, no contents, etc...}
requests_mock_datadir_missing_all = requests_mock_datadir_factory(
    ignore_urls=[
        "https://files.pythonhosted.org/packages/ec/65/c0116953c9a3f47de89e71964d6c7b0c783b01f29fa3390584dbf3046b4d/0805nexter-1.1.0.zip",  # noqa
        "https://files.pythonhosted.org/packages/c4/a0/4562cda161dc4ecbbe9e2a11eb365400c0461845c5be70d73869786809c4/0805nexter-1.2.0.zip",  # noqa
    ]
)


def test_no_release_artifact(swh_config, requests_mock_datadir_missing_all):
    """Load a pypi project with all artifacts missing ends up with no snapshot

    """
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "uneventful"
    assert actual_load_status["snapshot_id"] is not None

    stats = get_stats(loader.storage)
    assert {
        "content": 0,
        "directory": 0,
        "origin": 1,
        "origin_visit": 1,
        "person": 0,
        "release": 0,
        "revision": 0,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    assert_last_visit_matches(loader.storage, url, status="partial", type="pypi")


# problem during loading:
# {visit: partial, status: uneventful, no snapshot}


def test_release_with_traceback(swh_config, requests_mock_datadir):
    url = "https://pypi.org/project/0805nexter"
    with patch(
        "swh.loader.package.pypi.loader.PyPILoader.last_snapshot",
        side_effect=ValueError("Fake problem to fail the visit"),
    ):
        loader = PyPILoader(url)

        actual_load_status = loader.load()
        assert actual_load_status == {"status": "failed"}

        stats = get_stats(loader.storage)

        assert {
            "content": 0,
            "directory": 0,
            "origin": 1,
            "origin_visit": 1,
            "person": 0,
            "release": 0,
            "revision": 0,
            "skipped_content": 0,
            "snapshot": 0,
        } == stats

        assert_last_visit_matches(loader.storage, url, status="partial", type="pypi")


# problem during loading: failure early enough in between swh contents...
# some contents (contents, directories, etc...) have been written in storage
# {visit: partial, status: eventful, no snapshot}

# problem during loading: failure late enough we can have snapshots (some
# revisions are written in storage already)
# {visit: partial, status: eventful, snapshot}

# "normal" cases (for the same origin) #


requests_mock_datadir_missing_one = requests_mock_datadir_factory(
    ignore_urls=[
        "https://files.pythonhosted.org/packages/ec/65/c0116953c9a3f47de89e71964d6c7b0c783b01f29fa3390584dbf3046b4d/0805nexter-1.1.0.zip",  # noqa
    ]
)

# some missing release artifacts:
# {visit partial, status: eventful, 1 snapshot}


def test_revision_metadata_structure(swh_config, requests_mock_datadir):
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None

    expected_revision_id = hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21")
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(
        revision["metadata"],
        paths=[
            ("intrinsic.tool", str),
            ("intrinsic.raw", dict),
            ("extrinsic.provider", str),
            ("extrinsic.when", str),
            ("extrinsic.raw", dict),
            ("original_artifact", list),
        ],
    )

    for original_artifact in revision["metadata"]["original_artifact"]:
        check_metadata_paths(
            original_artifact,
            paths=[("filename", str), ("length", int), ("checksums", dict),],
        )


def test_visit_with_missing_artifact(swh_config, requests_mock_datadir_missing_one):
    """Load a pypi project with some missing artifacts ends up with 1 snapshot

    """
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = "dd0e4201a232b1c104433741dbf45895b8ac9355"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    stats = get_stats(loader.storage)

    assert {
        "content": 3,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_contents = map(
        hash_to_bytes,
        [
            "405859113963cb7a797642b45f171d6360425d16",
            "e5686aa568fdb1d19d7f1329267082fe40482d31",
            "83ecf6ec1114fd260ca7a833a2d165e71258c338",
        ],
    )

    assert list(loader.storage.content_missing_per_sha1(expected_contents)) == []

    expected_dirs = map(
        hash_to_bytes,
        [
            "b178b66bd22383d5f16f4f5c923d39ca798861b4",
            "c3a58f8b57433a4b56caaa5033ae2e0931405338",
        ],
    )

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"): hash_to_bytes(
            "b178b66bd22383d5f16f4f5c923d39ca798861b4"
        ),  # noqa
    }
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        b"releases/1.2.0": {
            "target": hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"),
            "target_type": "revision",
        },
        b"HEAD": {"target": b"releases/1.2.0", "target_type": "alias",},
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    check_snapshot(expected_snapshot, storage=loader.storage)

    assert_last_visit_matches(loader.storage, url, status="partial", type="pypi")


def test_visit_with_1_release_artifact(swh_config, requests_mock_datadir):
    """With no prior visit, load a pypi project ends up with 1 snapshot

    """
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = "ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    stats = get_stats(loader.storage)
    assert {
        "content": 6,
        "directory": 4,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_contents = map(
        hash_to_bytes,
        [
            "a61e24cdfdab3bb7817f6be85d37a3e666b34566",
            "938c33483285fd8ad57f15497f538320df82aeb8",
            "a27576d60e08c94a05006d2e6d540c0fdb5f38c8",
            "405859113963cb7a797642b45f171d6360425d16",
            "e5686aa568fdb1d19d7f1329267082fe40482d31",
            "83ecf6ec1114fd260ca7a833a2d165e71258c338",
        ],
    )

    assert list(loader.storage.content_missing_per_sha1(expected_contents)) == []

    expected_dirs = map(
        hash_to_bytes,
        [
            "05219ba38bc542d4345d5638af1ed56c7d43ca7d",
            "cf019eb456cf6f78d8c4674596f1c9a97ece8f44",
            "b178b66bd22383d5f16f4f5c923d39ca798861b4",
            "c3a58f8b57433a4b56caaa5033ae2e0931405338",
        ],
    )

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"): hash_to_bytes(
            "05219ba38bc542d4345d5638af1ed56c7d43ca7d"
        ),  # noqa
        hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"): hash_to_bytes(
            "b178b66bd22383d5f16f4f5c923d39ca798861b4"
        ),  # noqa
    }
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        b"releases/1.1.0": {
            "target": hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"),
            "target_type": "revision",
        },
        b"releases/1.2.0": {
            "target": hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"),
            "target_type": "revision",
        },
        b"HEAD": {"target": b"releases/1.2.0", "target_type": "alias",},
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")


def test_multiple_visits_with_no_change(swh_config, requests_mock_datadir):
    """Multiple visits with no changes results in 1 same snapshot

    """
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    snapshot_id = "ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": snapshot_id,
    }
    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")

    stats = get_stats(loader.storage)

    assert {
        "content": 6,
        "directory": 4,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_branches = {
        b"releases/1.1.0": {
            "target": hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"),
            "target_type": "revision",
        },
        b"releases/1.2.0": {
            "target": hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"),
            "target_type": "revision",
        },
        b"HEAD": {"target": b"releases/1.2.0", "target_type": "alias",},
    }

    expected_snapshot = {
        "id": hash_to_bytes(snapshot_id),
        "branches": expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    actual_load_status2 = loader.load()
    assert actual_load_status2 == {
        "status": "uneventful",
        "snapshot_id": actual_load_status2["snapshot_id"],
    }

    visit_status2 = assert_last_visit_matches(
        loader.storage, url, status="full", type="pypi"
    )

    stats2 = get_stats(loader.storage)
    expected_stats2 = stats.copy()
    expected_stats2["origin_visit"] = 1 + 1
    assert expected_stats2 == stats2

    # same snapshot
    actual_snapshot_id = visit_status2.snapshot
    assert actual_snapshot_id == hash_to_bytes(snapshot_id)


def test_incremental_visit(swh_config, requests_mock_datadir_visits):
    """With prior visit, 2nd load will result with a different snapshot

    """
    url = "https://pypi.org/project/0805nexter"
    loader = PyPILoader(url)

    visit1_actual_load_status = loader.load()
    visit1_stats = get_stats(loader.storage)
    expected_snapshot_id = "ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a"
    assert visit1_actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")

    assert {
        "content": 6,
        "directory": 4,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    } == visit1_stats

    # Reset internal state
    loader._info = None

    visit2_actual_load_status = loader.load()
    visit2_stats = get_stats(loader.storage)

    assert visit2_actual_load_status["status"] == "eventful"
    expected_snapshot_id2 = "2e5149a7b0725d18231a37b342e9b7c4e121f283"
    assert visit2_actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id2,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")

    assert {
        "content": 6 + 1,  # 1 more content
        "directory": 4 + 2,  # 2 more directories
        "origin": 1,
        "origin_visit": 1 + 1,
        "person": 1,
        "release": 0,
        "revision": 2 + 1,  # 1 more revision
        "skipped_content": 0,
        "snapshot": 1 + 1,  # 1 more snapshot
    } == visit2_stats

    expected_contents = map(
        hash_to_bytes,
        [
            "a61e24cdfdab3bb7817f6be85d37a3e666b34566",
            "938c33483285fd8ad57f15497f538320df82aeb8",
            "a27576d60e08c94a05006d2e6d540c0fdb5f38c8",
            "405859113963cb7a797642b45f171d6360425d16",
            "e5686aa568fdb1d19d7f1329267082fe40482d31",
            "83ecf6ec1114fd260ca7a833a2d165e71258c338",
            "92689fa2b7fb4d4fc6fb195bf73a50c87c030639",
        ],
    )

    assert list(loader.storage.content_missing_per_sha1(expected_contents)) == []

    expected_dirs = map(
        hash_to_bytes,
        [
            "05219ba38bc542d4345d5638af1ed56c7d43ca7d",
            "cf019eb456cf6f78d8c4674596f1c9a97ece8f44",
            "b178b66bd22383d5f16f4f5c923d39ca798861b4",
            "c3a58f8b57433a4b56caaa5033ae2e0931405338",
            "e226e7e4ad03b4fc1403d69a18ebdd6f2edd2b3a",
            "52604d46843b898f5a43208045d09fcf8731631b",
        ],
    )

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"): hash_to_bytes(
            "05219ba38bc542d4345d5638af1ed56c7d43ca7d"
        ),  # noqa
        hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"): hash_to_bytes(
            "b178b66bd22383d5f16f4f5c923d39ca798861b4"
        ),  # noqa
        hash_to_bytes("51247143b01445c9348afa9edfae31bf7c5d86b1"): hash_to_bytes(
            "e226e7e4ad03b4fc1403d69a18ebdd6f2edd2b3a"
        ),  # noqa
    }

    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        b"releases/1.1.0": {
            "target": hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"),
            "target_type": "revision",
        },
        b"releases/1.2.0": {
            "target": hash_to_bytes("e445da4da22b31bfebb6ffc4383dbf839a074d21"),
            "target_type": "revision",
        },
        b"releases/1.3.0": {
            "target": hash_to_bytes("51247143b01445c9348afa9edfae31bf7c5d86b1"),
            "target_type": "revision",
        },
        b"HEAD": {"target": b"releases/1.3.0", "target_type": "alias",},
    }
    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id2),
        "branches": expected_branches,
    }

    check_snapshot(expected_snapshot, loader.storage)

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")

    urls = [
        m.url
        for m in requests_mock_datadir_visits.request_history
        if m.url.startswith("https://files.pythonhosted.org")
    ]
    # visited each artifact once across 2 visits
    assert len(urls) == len(set(urls))


# release artifact, no new artifact
# {visit full, status uneventful, same snapshot as before}

# release artifact, old artifact with different checksums
# {visit full, status full, new snapshot with shared history and some new
# different history}

# release with multiple sdist artifacts per pypi "version"
# snapshot branch output is different


def test_visit_1_release_with_2_artifacts(swh_config, requests_mock_datadir):
    """With no prior visit, load a pypi project ends up with 1 snapshot

    """
    url = "https://pypi.org/project/nexter"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = "a27e638a4dad6fbfa273c6ebec1c4bf320fb84c6"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    expected_branches = {
        b"releases/1.1.0/nexter-1.1.0.zip": {
            "target": hash_to_bytes("4c99891f93b81450385777235a37b5e966dd1571"),
            "target_type": "revision",
        },
        b"releases/1.1.0/nexter-1.1.0.tar.gz": {
            "target": hash_to_bytes("0bf88f5760cca7665d0af4d6575d9301134fe11a"),
            "target_type": "revision",
        },
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")


def test_pypi_artifact_to_revision_id_none():
    """Current loader version should stop soon if nothing can be found

    """
    artifact_metadata = {
        "digests": {
            "sha256": "6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec",  # noqa
        },
    }

    assert artifact_to_revision_id({}, artifact_metadata) is None

    known_artifacts = {
        "b11ebac8c9d0c9e5063a2df693a18e3aba4b2f92": {
            "original_artifact": {"sha256": "something-irrelevant",},
        },
    }

    assert artifact_to_revision_id(known_artifacts, artifact_metadata) is None


def test_pypi_artifact_to_revision_id_old_loader_version():
    """Current loader version should solve old metadata scheme

    """
    artifact_metadata = {
        "digests": {
            "sha256": "6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec",  # noqa
        }
    }

    known_artifacts = {
        hash_to_bytes("b11ebac8c9d0c9e5063a2df693a18e3aba4b2f92"): {
            "original_artifact": {"sha256": "something-wrong",},
        },
        hash_to_bytes("845673bfe8cbd31b1eaf757745a964137e6f9116"): {
            "original_artifact": {
                "sha256": "6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec",  # noqa
            },
        },
    }

    assert artifact_to_revision_id(known_artifacts, artifact_metadata) == hash_to_bytes(
        "845673bfe8cbd31b1eaf757745a964137e6f9116"
    )


def test_pypi_artifact_to_revision_id_current_loader_version():
    """Current loader version should be able to solve current metadata scheme

    """
    artifact_metadata = {
        "digests": {
            "sha256": "6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec",  # noqa
        }
    }

    known_artifacts = {
        hash_to_bytes("b11ebac8c9d0c9e5063a2df693a18e3aba4b2f92"): {
            "original_artifact": [
                {
                    "checksums": {
                        "sha256": "6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec",  # noqa
                    },
                }
            ],
        },
        hash_to_bytes("845673bfe8cbd31b1eaf757745a964137e6f9116"): {
            "original_artifact": [{"checksums": {"sha256": "something-wrong"},}],
        },
    }

    assert artifact_to_revision_id(known_artifacts, artifact_metadata) == hash_to_bytes(
        "b11ebac8c9d0c9e5063a2df693a18e3aba4b2f92"
    )


def test_pypi_artifact_to_revision_id_failures():
    with pytest.raises(KeyError, match="sha256"):
        artifact_metadata = {
            "digests": {},
        }
        assert artifact_to_revision_id({}, artifact_metadata)

    with pytest.raises(KeyError, match="digests"):
        artifact_metadata = {
            "something": "wrong",
        }
        assert artifact_to_revision_id({}, artifact_metadata)


def test_pypi_artifact_with_no_intrinsic_metadata(swh_config, requests_mock_datadir):
    """Skip artifact with no intrinsic metadata during ingestion

    """
    url = "https://pypi.org/project/upymenu"
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = "1a8893e6a86f444e8be8e7bda6cb34fb1735a00e"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    # no branch as one artifact without any intrinsic metadata
    expected_snapshot = {"id": hash_to_bytes(expected_snapshot_id), "branches": {}}
    check_snapshot(expected_snapshot, loader.storage)

    assert_last_visit_matches(loader.storage, url, status="full", type="pypi")



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib
import logging
import os

from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, Optional, Tuple

from swh.core import config
from swh.model.model import (
    BaseContent,
    Content,
    SkippedContent,
    Directory,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Revision,
    Release,
    Sha1Git,
    Snapshot,
)
from swh.storage import get_storage
from swh.storage.utils import now


class BaseLoader(config.SWHConfig, metaclass=ABCMeta):
    """Mixin base class for loader.

    To use this class, you must:

    - inherit from this class
    - and implement the @abstractmethod methods:

      - :func:`prepare`: First step executed by the loader to prepare some
        state needed by the `func`:load method.

      - :func:`get_origin`: Retrieve the origin that is currently being loaded.

      - :func:`fetch_data`: Fetch the data is actually the method to implement
        to compute data to inject in swh (through the store_data method)

      - :func:`store_data`: Store data fetched.

      - :func:`visit_status`: Explicit status of the visit ('partial' or
        'full')

      - :func:`load_status`: Explicit status of the loading, for use by the
        scheduler (eventful/uneventful/temporary failure/permanent failure).

      - :func:`cleanup`: Last step executed by the loader.

    The entry point for the resulting loader is :func:`load`.

    You can take a look at some example classes:

    - :class:`BaseSvnLoader`

    """

    CONFIG_BASE_FILENAME = None  # type: Optional[str]

    DEFAULT_CONFIG = {
        "storage": ("dict", {"cls": "remote", "url": "http://localhost:5002/",}),
        "max_content_size": ("int", 100 * 1024 * 1024),
        "save_data": ("bool", False),
        "save_data_path": ("str", ""),
    }  # type: Dict[str, Tuple[str, Any]]

    ADDITIONAL_CONFIG = {}  # type: Dict[str, Tuple[str, Any]]

    def __init__(
        self, logging_class: Optional[str] = None, config: Dict[str, Any] = {}
    ):
        if config:
            self.config = config
        else:
            self.config = self.parse_config_file(
                additional_configs=[self.ADDITIONAL_CONFIG]
            )

        self.storage = get_storage(**self.config["storage"])

        if logging_class is None:
            logging_class = "%s.%s" % (
                self.__class__.__module__,
                self.__class__.__name__,
            )
        self.log = logging.getLogger(logging_class)

        _log = logging.getLogger("requests.packages.urllib3.connectionpool")
        _log.setLevel(logging.WARN)

        self.max_content_size = self.config["max_content_size"]

        # possibly overridden in self.prepare method
        self.visit_date: Optional[datetime.datetime] = None

        self.origin: Optional[Origin] = None

        if not hasattr(self, "visit_type"):
            self.visit_type: Optional[str] = None

        self.origin_metadata: Dict[str, Any] = {}

        self.loaded_snapshot_id: Optional[Sha1Git] = None

        # Make sure the config is sane
        save_data = self.config.get("save_data")
        if save_data:
            path = self.config["save_data_path"]
            os.stat(path)
            if not os.access(path, os.R_OK | os.W_OK):
                raise PermissionError("Permission denied: %r" % path)

    def save_data(self) -> None:
        """Save the data associated to the current load"""
        raise NotImplementedError

    def get_save_data_path(self) -> str:
        """The path to which we archive the loader's raw data"""
        if not hasattr(self, "__save_data_path"):
            year = str(self.visit_date.year)  # type: ignore

            assert self.origin
            url = self.origin.url.encode("utf-8")
            origin_url_hash = hashlib.sha1(url).hexdigest()

            path = "%s/sha1:%s/%s/%s" % (
                self.config["save_data_path"],
                origin_url_hash[0:2],
                origin_url_hash,
                year,
            )

            os.makedirs(path, exist_ok=True)
            self.__save_data_path = path

        return self.__save_data_path

    def flush(self) -> None:
        """Flush any potential buffered data not sent to swh-storage.

        """
        self.storage.flush()

    @abstractmethod
    def cleanup(self) -> None:
        """Last step executed by the loader.

        """
        pass

    @abstractmethod
    def prepare_origin_visit(self, *args, **kwargs) -> None:
        """First step executed by the loader to prepare origin and visit
           references. Set/update self.origin, and
           optionally self.origin_url, self.visit_date.

        """
        pass

    def _store_origin_visit(self) -> None:
        """Store origin and visit references. Sets the self.visit references.

        """
        assert self.origin
        self.storage.origin_add([self.origin])

        if not self.visit_date:  # now as default visit_date if not provided
            self.visit_date = datetime.datetime.now(tz=datetime.timezone.utc)
        assert isinstance(self.visit_date, datetime.datetime)
        assert isinstance(self.visit_type, str)
        self.visit = self.storage.origin_visit_add(
            [
                OriginVisit(
                    origin=self.origin.url, date=self.visit_date, type=self.visit_type,
                )
            ]
        )[0]

    @abstractmethod
    def prepare(self, *args, **kwargs) -> None:
        """Second step executed by the loader to prepare some state needed by
           the loader.

        """
        pass

    def get_origin(self) -> Origin:
        """Get the origin that is currently being loaded.
        self.origin should be set in :func:`prepare_origin`

        Returns:
          dict: an origin ready to be sent to storage by
          :func:`origin_add_one`.
        """
        assert self.origin
        return self.origin

    @abstractmethod
    def fetch_data(self) -> bool:
        """Fetch the data from the source the loader is currently loading
           (ex: git/hg/svn/... repository).

        Returns:
            a value that is interpreted as a boolean. If True, fetch_data needs
            to be called again to complete loading.

        """
        pass

    @abstractmethod
    def store_data(self):
        """Store fetched data in the database.

        Should call the :func:`maybe_load_xyz` methods, which handle the
        bundles sent to storage, rather than send directly.
        """
        pass

    def store_metadata(self) -> None:
        """Store fetched metadata in the database.

        For more information, see implementation in :class:`DepositLoader`.
        """
        pass

    def load_status(self) -> Dict[str, str]:
        """Detailed loading status.

        Defaults to logging an eventful load.

        Returns: a dictionary that is eventually passed back as the task's
          result to the scheduler, allowing tuning of the task recurrence
          mechanism.
        """
        return {
            "status": "eventful",
        }

    def post_load(self, success: bool = True) -> None:
        """Permit the loader to do some additional actions according to status
        after the loading is done. The flag success indicates the
        loading's status.

        Defaults to doing nothing.

        This is up to the implementer of this method to make sure this
        does not break.

        Args:
            success (bool): the success status of the loading

        """
        pass

    def visit_status(self) -> str:
        """Detailed visit status.

        Defaults to logging a full visit.
        """
        return "full"

    def pre_cleanup(self) -> None:
        """As a first step, will try and check for dangling data to cleanup.
        This should do its best to avoid raising issues.

        """
        pass

    def load(self, *args, **kwargs) -> Dict[str, str]:
        r"""Loading logic for the loader to follow:

        - 1. Call :meth:`prepare_origin_visit` to prepare the
             origin and visit we will associate loading data to
        - 2. Store the actual ``origin_visit`` to storage
        - 3. Call :meth:`prepare` to prepare any eventual state
        - 4. Call :meth:`get_origin` to get the origin we work with and store

        - while True:

          - 5. Call :meth:`fetch_data` to fetch the data to store
          - 6. Call :meth:`store_data` to store the data

        - 7. Call :meth:`cleanup` to clean up any eventual state put in place
             in :meth:`prepare` method.

        """
        try:
            self.pre_cleanup()
        except Exception:
            msg = "Cleaning up dangling data failed! Continue loading."
            self.log.warning(msg)

        self.prepare_origin_visit(*args, **kwargs)
        self._store_origin_visit()

        assert self.origin

        try:
            self.prepare(*args, **kwargs)

            while True:
                more_data_to_fetch = self.fetch_data()
                self.store_data()
                if not more_data_to_fetch:
                    break

            self.store_metadata()
            visit_status = OriginVisitStatus(
                origin=self.origin.url,
                visit=self.visit.visit,
                date=now(),
                status=self.visit_status(),
                snapshot=self.loaded_snapshot_id,
            )
            self.storage.origin_visit_status_add([visit_status])
            self.post_load()
        except Exception:
            self.log.exception(
                "Loading failure, updating to `partial` status",
                extra={"swh_task_args": args, "swh_task_kwargs": kwargs,},
            )
            visit_status = OriginVisitStatus(
                origin=self.origin.url,
                visit=self.visit.visit,
                date=now(),
                status="partial",
                snapshot=self.loaded_snapshot_id,
            )
            self.storage.origin_visit_status_add([visit_status])
            self.post_load(success=False)
            return {"status": "failed"}
        finally:
            self.flush()
            self.cleanup()

        return self.load_status()


class DVCSLoader(BaseLoader):
    """This base class is a pattern for dvcs loaders (e.g. git, mercurial).

    Those loaders are able to load all the data in one go. For example, the
    loader defined in swh-loader-git :class:`BulkUpdater`.

    For other loaders (stateful one, (e.g :class:`SWHSvnLoader`),
    inherit directly from :class:`BaseLoader`.

    """

    ADDITIONAL_CONFIG = {}  # type: Dict[str, Tuple[str, Any]]

    def cleanup(self) -> None:
        """Clean up an eventual state installed for computations."""
        pass

    def has_contents(self) -> bool:
        """Checks whether we need to load contents"""
        return True

    def get_contents(self) -> Iterable[BaseContent]:
        """Get the contents that need to be loaded"""
        raise NotImplementedError

    def has_directories(self) -> bool:
        """Checks whether we need to load directories"""
        return True

    def get_directories(self) -> Iterable[Directory]:
        """Get the directories that need to be loaded"""
        raise NotImplementedError

    def has_revisions(self) -> bool:
        """Checks whether we need to load revisions"""
        return True

    def get_revisions(self) -> Iterable[Revision]:
        """Get the revisions that need to be loaded"""
        raise NotImplementedError

    def has_releases(self) -> bool:
        """Checks whether we need to load releases"""
        return True

    def get_releases(self) -> Iterable[Release]:
        """Get the releases that need to be loaded"""
        raise NotImplementedError

    def get_snapshot(self) -> Snapshot:
        """Get the snapshot that needs to be loaded"""
        raise NotImplementedError

    def eventful(self) -> bool:
        """Whether the load was eventful"""
        raise NotImplementedError

    def store_data(self) -> None:
        assert self.origin
        if self.config.get("save_data"):
            self.save_data()

        if self.has_contents():
            contents = []
            skipped_contents = []
            for obj in self.get_contents():
                if isinstance(obj, Content):
                    contents.append(obj)
                elif isinstance(obj, SkippedContent):
                    skipped_contents.append(obj)
                else:
                    raise TypeError(f"Unexpected content type: {obj}")
            self.storage.skipped_content_add(skipped_contents)
            self.storage.content_add(contents)
        if self.has_directories():
            self.storage.directory_add(self.get_directories())
        if self.has_revisions():
            self.storage.revision_add(self.get_revisions())
        if self.has_releases():
            self.storage.release_add(self.get_releases())
        snapshot = self.get_snapshot()
        self.storage.snapshot_add([snapshot])
        self.flush()
        self.loaded_snapshot_id = snapshot.id



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import re

import pytest

from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.loader.package.deposit.loader import DepositLoader

from swh.loader.package.tests.common import check_metadata_paths
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
)

from swh.core.pytest_plugin import requests_mock_datadir_factory


DEPOSIT_URL = "https://deposit.softwareheritage.org/1/private"


@pytest.fixture
def requests_mock_datadir(requests_mock_datadir):
    """Enhance default mock data to mock put requests as the loader does some
       internal update queries there.

    """
    requests_mock_datadir.put(re.compile("https"))
    return requests_mock_datadir


def test_deposit_init_ok(swh_config, swh_loader_config):
    url = "some-url"
    deposit_id = 999
    loader = DepositLoader(url, deposit_id)  # Something that does not exist

    assert loader.url == url
    assert loader.client is not None
    assert loader.client.base_url == swh_loader_config["deposit"]["url"]


def test_deposit_loading_unknown_deposit(swh_config, requests_mock_datadir):
    """Loading an unknown deposit should fail

    no origin, no visit, no snapshot
    """
    # private api url form: 'https://deposit.s.o/1/private/hal/666/raw/'
    url = "some-url"
    unknown_deposit_id = 667
    loader = DepositLoader(url, unknown_deposit_id)  # does not exist

    actual_load_status = loader.load()
    assert actual_load_status == {"status": "failed"}

    stats = get_stats(loader.storage)

    assert {
        "content": 0,
        "directory": 0,
        "origin": 0,
        "origin_visit": 0,
        "person": 0,
        "release": 0,
        "revision": 0,
        "skipped_content": 0,
        "snapshot": 0,
    } == stats


requests_mock_datadir_missing_one = requests_mock_datadir_factory(
    ignore_urls=[f"{DEPOSIT_URL}/666/raw/",]
)


def test_deposit_loading_failure_to_retrieve_1_artifact(
    swh_config, requests_mock_datadir_missing_one
):
    """Deposit with missing artifact ends up with an uneventful/partial visit

    """
    # private api url form: 'https://deposit.s.o/1/private/hal/666/raw/'
    url = "some-url-2"
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "uneventful"
    assert actual_load_status["snapshot_id"] is not None

    assert_last_visit_matches(loader.storage, url, status="partial", type="deposit")

    stats = get_stats(loader.storage)
    assert {
        "content": 0,
        "directory": 0,
        "origin": 1,
        "origin_visit": 1,
        "person": 0,
        "release": 0,
        "revision": 0,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats


def test_revision_metadata_structure(swh_config, requests_mock_datadir):
    url = "https://hal-test.archives-ouvertes.fr/some-external-id"
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None
    expected_revision_id = hash_to_bytes("637318680351f5d78856d13264faebbd91efe9bb")
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(
        revision["metadata"],
        paths=[
            ("extrinsic.provider", str),
            ("extrinsic.when", str),
            ("extrinsic.raw", dict),
            ("original_artifact", list),
        ],
    )

    # Only 2 top-level keys now
    assert set(revision["metadata"].keys()) == {"extrinsic", "original_artifact"}

    for original_artifact in revision["metadata"]["original_artifact"]:
        check_metadata_paths(
            original_artifact,
            paths=[("filename", str), ("length", int), ("checksums", dict),],
        )


def test_deposit_loading_ok(swh_config, requests_mock_datadir):
    url = "https://hal-test.archives-ouvertes.fr/some-external-id"
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    expected_snapshot_id = "b2b327b33dc85818bd23c3ccda8b7e675a66ecbd"
    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    assert_last_visit_matches(loader.storage, url, status="full", type="deposit")

    stats = get_stats(loader.storage)
    assert {
        "content": 303,
        "directory": 12,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    revision_id = "637318680351f5d78856d13264faebbd91efe9bb"
    expected_branches = {
        b"HEAD": {"target": hash_to_bytes(revision_id), "target_type": "revision",},
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    check_snapshot(expected_snapshot, storage=loader.storage)

    revision = next(loader.storage.revision_get([hash_to_bytes(revision_id)]))
    assert revision

    # check metadata

    fetcher = {
        "name": "swh-deposit",
        "version": "0.0.1",
    }

    authority = {
        "type": "deposit_client",
        "url": "https://hal-test.archives-ouvertes.fr/",
    }

    orig_meta = loader.storage.origin_metadata_get(url, authority)
    assert orig_meta is not None
    assert isinstance(orig_meta, dict)
    assert len(orig_meta["results"]) == 1
    assert orig_meta["next_page_token"] is None
    orig_meta0 = orig_meta["results"][0]

    assert orig_meta0["authority"] == authority
    assert orig_meta0["fetcher"] == fetcher

    # Retrieve the information for deposit status update query to the deposit
    urls = [
        m
        for m in requests_mock_datadir.request_history
        if m.url == f"{DEPOSIT_URL}/{deposit_id}/update/"
    ]

    assert len(urls) == 1
    update_query = urls[0]

    body = update_query.json()
    expected_body = {
        "status": "done",
        "revision_id": revision_id,
        "directory_id": hash_to_hex(revision["directory"]),
        "snapshot_id": expected_snapshot_id,
        "origin_url": url,
    }

    assert body == expected_body


def test_deposit_loading_ok_2(swh_config, requests_mock_datadir):
    """Field dates should be se appropriately

    """
    external_id = "some-external-id"
    url = f"https://hal-test.archives-ouvertes.fr/{external_id}"
    deposit_id = 777
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    expected_snapshot_id = "3e68440fdd7c81d283f8f3aebb6f0c8657864192"

    assert actual_load_status == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }
    assert_last_visit_matches(loader.storage, url, status="full", type="deposit")

    revision_id = "564d18943d71be80d0d73b43a77cfb205bcde96c"
    expected_branches = {
        b"HEAD": {"target": hash_to_bytes(revision_id), "target_type": "revision"}
    }
    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }

    check_snapshot(expected_snapshot, storage=loader.storage)

    raw_meta = loader.client.metadata_get(deposit_id)
    # Ensure the date fields are set appropriately in the revision

    # Retrieve the revision
    revision = next(loader.storage.revision_get([hash_to_bytes(revision_id)]))
    assert revision
    assert revision["date"] == raw_meta["deposit"]["author_date"]
    assert revision["committer_date"] == raw_meta["deposit"]["committer_date"]

    read_api = f"{DEPOSIT_URL}/{deposit_id}/meta/"

    assert revision["metadata"] == {
        "extrinsic": {
            "provider": read_api,
            "raw": {
                "origin": {"type": "deposit", "url": url,},
                "origin_metadata": {
                    "metadata": {
                        "@xmlns": ["http://www.w3.org/2005/Atom"],
                        "author": ["some awesome author", "another one", "no one",],
                        "codemeta:dateCreated": "2017-10-07T15:17:08Z",
                        "codemeta:datePublished": "2017-10-08T15:00:00Z",
                        "external_identifier": "some-external-id",
                        "url": url,
                    },
                    "provider": {
                        "metadata": None,
                        "provider_name": "hal",
                        "provider_type": "deposit_client",
                        "provider_url": "https://hal-test.archives-ouvertes.fr/",
                    },
                    "tool": {
                        "configuration": {"sword_version": "2"},
                        "name": "swh-deposit",
                        "version": "0.0.1",
                    },
                },
            },
            "when": revision["metadata"]["extrinsic"]["when"],  # dynamic
        },
        "original_artifact": [
            {
                "checksums": {
                    "sha1": "f8c63d7c890a7453498e6cf9fef215d85ec6801d",
                    "sha256": "474bf646aeeff6d945eb752b1a9f8a40f3d81a88909ee7bd2d08cc822aa361e6",  # noqa
                },
                "filename": "archive.zip",
                "length": 956830,
            }
        ],
    }

    fetcher = {
        "name": "swh-deposit",
        "version": "0.0.1",
    }

    authority = {
        "type": "deposit_client",
        "url": "https://hal-test.archives-ouvertes.fr/",
    }

    # Check the metadata swh side
    orig_meta = loader.storage.origin_metadata_get(url, authority)
    assert orig_meta is not None
    assert isinstance(orig_meta, dict)
    assert len(orig_meta["results"]) == 1
    assert orig_meta["next_page_token"] is None

    assert len(orig_meta["results"]) == 1

    orig_meta0 = orig_meta["results"][0]
    # dynamic, a pain to display and not that interesting
    orig_meta0.pop("discovery_date")

    assert orig_meta0 == {
        "origin_url": "https://hal-test.archives-ouvertes.fr/some-external-id",
        "metadata": json.dumps(
            {
                "@xmlns": ["http://www.w3.org/2005/Atom"],
                "author": ["some awesome author", "another one", "no one"],
                "codemeta:dateCreated": "2017-10-07T15:17:08Z",
                "codemeta:datePublished": "2017-10-08T15:00:00Z",
                "external_identifier": "some-external-id",
                "url": "https://hal-test.archives-ouvertes.fr/some-external-id",
            }
        ).encode(),
        "format": "sword-v2-atom-codemeta-v2-in-json",
        "authority": authority,
        "fetcher": fetcher,
    }

    # Retrieve the information for deposit status update query to the deposit
    urls = [
        m
        for m in requests_mock_datadir.request_history
        if m.url == f"{DEPOSIT_URL}/{deposit_id}/update/"
    ]

    assert len(urls) == 1
    update_query = urls[0]

    body = update_query.json()
    expected_body = {
        "status": "done",
        "revision_id": revision_id,
        "directory_id": hash_to_hex(revision["directory"]),
        "snapshot_id": expected_snapshot_id,
        "origin_url": url,
    }

    assert body == expected_body



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import pytest

import os
import subprocess

from swh.model.from_disk import DentryPerms
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ObjectType,
    OriginVisit,
    OriginVisitStatus,
    Person,
    Release,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    TargetType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.hashutil import hash_to_bytes

from swh.loader.tests import (
    assert_last_visit_matches,
    encode_target,
    check_snapshot,
    prepare_repository_from_archive,
    InconsistentAliasBranchError,
    InexistentObjectsError,
)


hash_hex = "43e45d56f88993aae6a0198013efa80716fd8920"


ORIGIN_VISIT = OriginVisit(
    origin="some-url",
    visit=1,
    date=datetime.datetime.now(tz=datetime.timezone.utc),
    type="archive",
)


ORIGIN_VISIT_STATUS = OriginVisitStatus(
    origin="some-url",
    visit=1,
    date=datetime.datetime.now(tz=datetime.timezone.utc),
    status="full",
    snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    metadata=None,
)


CONTENT = Content(
    data=b"42\n",
    length=3,
    sha1=hash_to_bytes("34973274ccef6ab4dfaaf86599792fa9c3fe4689"),
    sha1_git=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    sha256=hash_to_bytes(
        "673650f936cb3b0a2f93ce09d81be10748b1b203c19e8176b4eefc1964a0cf3a"
    ),
    blake2s256=hash_to_bytes(
        "d5fe1939576527e42cfd76a9455a2432fe7f56669564577dd93c4280e76d661d"
    ),
    status="visible",
)


DIRECTORY = Directory(
    id=hash_to_bytes("34f335a750111ca0a8b64d8034faec9eedc396be"),
    entries=tuple(
        [
            DirectoryEntry(
                name=b"foo",
                type="file",
                target=CONTENT.sha1_git,
                perms=DentryPerms.content,
            )
        ]
    ),
)


REVISION = Revision(
    id=hash_to_bytes("066b1b62dbfa033362092af468bf6cfabec230e7"),
    message=b"hello",
    author=Person(
        name=b"Nicolas Dandrimont",
        email=b"nicolas@example.com",
        fullname=b"Nicolas Dandrimont <nicolas@example.com> ",
    ),
    date=TimestampWithTimezone(
        timestamp=Timestamp(seconds=1234567890, microseconds=0),
        offset=120,
        negative_utc=False,
    ),
    committer=Person(
        name=b"St\xc3fano Zacchiroli",
        email=b"stefano@example.com",
        fullname=b"St\xc3fano Zacchiroli <stefano@example.com>",
    ),
    committer_date=TimestampWithTimezone(
        timestamp=Timestamp(seconds=1123456789, microseconds=0),
        offset=0,
        negative_utc=True,
    ),
    parents=(),
    type=RevisionType.GIT,
    directory=DIRECTORY.id,
    metadata={
        "checksums": {"sha1": "tarball-sha1", "sha256": "tarball-sha256",},
        "signed-off-by": "some-dude",
    },
    extra_headers=(
        (b"gpgsig", b"test123"),
        (b"mergetag", b"foo\\bar"),
        (b"mergetag", b"\x22\xaf\x89\x80\x01\x00"),
    ),
    synthetic=True,
)


RELEASE = Release(
    id=hash_to_bytes("3e9050196aa288264f2a9d279d6abab8b158448b"),
    name=b"v0.0.2",
    author=Person(
        name=b"tony", email=b"tony@ardumont.fr", fullname=b"tony <tony@ardumont.fr>",
    ),
    date=TimestampWithTimezone(
        timestamp=Timestamp(seconds=1634336813, microseconds=0),
        offset=0,
        negative_utc=False,
    ),
    target=REVISION.id,
    target_type=ObjectType.REVISION,
    message=b"yet another synthetic release",
    synthetic=True,
)


SNAPSHOT = Snapshot(
    id=hash_to_bytes("2498dbf535f882bc7f9a18fb16c9ad27fda7bab7"),
    branches={
        b"release/0.1.0": SnapshotBranch(
            target=RELEASE.id, target_type=TargetType.RELEASE,
        ),
        b"HEAD": SnapshotBranch(target=REVISION.id, target_type=TargetType.REVISION,),
        b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
        b"evaluation": SnapshotBranch(  # branch dedicated to not exist in storage
            target=hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            target_type=TargetType.REVISION,
        ),
    },
)


@pytest.fixture
def mock_storage(mocker):
    mock_storage = mocker.patch("swh.loader.tests.origin_get_latest_visit_status")
    mock_storage.return_value = ORIGIN_VISIT, ORIGIN_VISIT_STATUS
    return mock_storage


def test_assert_last_visit_matches_raise(mock_storage, mocker):
    """Not finding origin visit_and_statu should raise

    """
    # overwrite so we raise because we do not find the right visit
    mock_storage.return_value = None

    with pytest.raises(AssertionError, match="Origin url has no visits"):
        assert_last_visit_matches(mock_storage, "url", status="full")

    assert mock_storage.called is True


def test_assert_last_visit_matches_wrong_status(mock_storage, mocker):
    """Wrong visit detected should raise AssertionError

    """
    expected_status = "partial"
    assert ORIGIN_VISIT_STATUS.status != expected_status
    with pytest.raises(AssertionError, match="Visit_status has status"):
        assert_last_visit_matches(mock_storage, "url", status=expected_status)

    assert mock_storage.called is True


def test_assert_last_visit_matches_wrong_type(mock_storage, mocker):
    """Wrong visit detected should raise AssertionError

    """
    expected_type = "git"
    assert ORIGIN_VISIT.type != expected_type
    with pytest.raises(AssertionError, match="Visit has type"):
        assert_last_visit_matches(
            mock_storage,
            "url",
            status=ORIGIN_VISIT_STATUS.status,
            type=expected_type,  # mismatched type will raise
        )

    assert mock_storage.called is True


def test_assert_last_visit_matches_wrong_snapshot(mock_storage, mocker):
    """Wrong visit detected should raise AssertionError

    """
    expected_snapshot_id = hash_to_bytes("e92cc0710eb6cf9efd5b920a8453e1e07157b6cd")
    assert ORIGIN_VISIT_STATUS.snapshot != expected_snapshot_id

    with pytest.raises(AssertionError, match="Visit_status points to snapshot"):
        assert_last_visit_matches(
            mock_storage,
            "url",
            status=ORIGIN_VISIT_STATUS.status,
            snapshot=expected_snapshot_id,  # mismatched snapshot will raise
        )

    assert mock_storage.called is True


def test_assert_last_visit_matches(mock_storage, mocker):
    """Correct visit detected should return the visit_status

    """
    visit_type = ORIGIN_VISIT.type
    visit_status = ORIGIN_VISIT_STATUS.status
    visit_snapshot = ORIGIN_VISIT_STATUS.snapshot

    actual_visit_status = assert_last_visit_matches(
        mock_storage,
        "url",
        type=visit_type,
        status=visit_status,
        snapshot=visit_snapshot,
    )

    assert actual_visit_status == ORIGIN_VISIT_STATUS
    assert mock_storage.called is True


def test_prepare_repository_from_archive_failure():
    # does not deal with inexistent archive so raise
    assert os.path.exists("unknown-archive") is False
    with pytest.raises(subprocess.CalledProcessError, match="exit status 2"):
        prepare_repository_from_archive("unknown-archive")


def test_prepare_repository_from_archive(datadir, tmp_path):
    archive_name = "0805nexter-1.1.0"
    archive_path = os.path.join(str(datadir), f"{archive_name}.tar.gz")
    assert os.path.exists(archive_path) is True

    tmp_path = str(tmp_path)  # deals with path string
    repo_url = prepare_repository_from_archive(
        archive_path, filename=archive_name, tmp_path=tmp_path
    )
    expected_uncompressed_archive_path = os.path.join(tmp_path, archive_name)
    assert repo_url == f"file://{expected_uncompressed_archive_path}"
    assert os.path.exists(expected_uncompressed_archive_path)


def test_prepare_repository_from_archive_no_filename(datadir, tmp_path):
    archive_name = "0805nexter-1.1.0"
    archive_path = os.path.join(str(datadir), f"{archive_name}.tar.gz")
    assert os.path.exists(archive_path) is True

    # deals with path as posix path (for tmp_path)
    repo_url = prepare_repository_from_archive(archive_path, tmp_path=tmp_path)

    tmp_path = str(tmp_path)
    expected_uncompressed_archive_path = os.path.join(tmp_path, archive_name)
    expected_repo_url = os.path.join(tmp_path, f"{archive_name}.tar.gz")
    assert repo_url == f"file://{expected_repo_url}"

    # passing along the filename does not influence the on-disk extraction
    # just the repo-url computation
    assert os.path.exists(expected_uncompressed_archive_path)


def test_encode_target():
    assert encode_target(None) is None

    for target_alias in ["something", b"something"]:
        target = {
            "target_type": "alias",
            "target": target_alias,
        }
        actual_alias_encode_target = encode_target(target)
        assert actual_alias_encode_target == {
            "target_type": "alias",
            "target": b"something",
        }

    for hash_ in [hash_hex, hash_to_bytes(hash_hex)]:
        target = {"target_type": "revision", "target": hash_}
        actual_encode_target = encode_target(target)
        assert actual_encode_target == {
            "target_type": "revision",
            "target": hash_to_bytes(hash_hex),
        }


def test_check_snapshot(swh_storage):
    """Everything should be fine when snapshot is found and the snapshot reference up to the
    revision exist in the storage.

    """
    # Create a consistent snapshot arborescence tree in storage
    found = False
    for entry in DIRECTORY.entries:
        if entry.target == CONTENT.sha1_git:
            found = True
            break
    assert found is True

    assert REVISION.directory == DIRECTORY.id
    assert RELEASE.target == REVISION.id

    for branch, target in SNAPSHOT.branches.items():
        if branch == b"alias":
            assert target.target in SNAPSHOT.branches
        elif branch == b"evaluation":
            # this one does not exist and we are safelisting its check below
            continue
        else:
            assert target.target in [REVISION.id, RELEASE.id]

    swh_storage.content_add([CONTENT])
    swh_storage.directory_add([DIRECTORY])
    swh_storage.revision_add([REVISION])
    swh_storage.release_add([RELEASE])
    s = swh_storage.snapshot_add([SNAPSHOT])
    assert s == {
        "snapshot:add": 1,
    }

    for snap in [SNAPSHOT, SNAPSHOT.to_dict()]:
        # all should be fine!
        check_snapshot(
            snap, swh_storage, allowed_empty=[(TargetType.REVISION, b"evaluation")]
        )


def test_check_snapshot_failures(swh_storage):
    """Failure scenarios:

    0. snapshot parameter is not a snapshot
    1. snapshot id is correct but branches mismatched
    2. snapshot id is not correct, it's not found in the storage
    3. snapshot reference an alias which does not exist
    4. snapshot is found in storage, targeted revision does not exist
    5. snapshot is found in storage, targeted revision exists but the directory the
       revision targets does not exist
    6. snapshot is found in storage, target revision exists, targeted directory by the
       revision exist. Content targeted by the directory does not exist.
    7. snapshot is found in storage, targeted release does not exist

    """
    snap_id_hex = "2498dbf535f882bc7f9a18fb16c9ad27fda7bab7"
    snapshot = Snapshot(
        id=hash_to_bytes(snap_id_hex),
        branches={
            b"master": SnapshotBranch(
                target=hash_to_bytes(hash_hex), target_type=TargetType.REVISION,
            ),
        },
    )

    s = swh_storage.snapshot_add([snapshot])
    assert s == {
        "snapshot:add": 1,
    }

    unexpected_snapshot = {
        # id is correct
        "id": hash_to_bytes(snap_id_hex),
        "branches": {
            b"master": {
                "target": hash_to_bytes(hash_hex),  # wrong branch
                "target_type": "release",
            }
        },
    }

    # 0. not a Snapshot object, raise!
    with pytest.raises(AssertionError, match="variable 'snapshot' must be a snapshot"):
        check_snapshot(ORIGIN_VISIT, swh_storage)

    # 1. snapshot id is correct but branches mismatched
    with pytest.raises(AssertionError, match="Differing attributes"):
        unexpected_snapshot["id"] = snapshot.id
        check_snapshot(unexpected_snapshot, swh_storage)

    # 2. snapshot id is not correct, it's not found in the storage
    wrong_snap_id = hash_to_bytes("999666f535f882bc7f9a18fb16c9ad27fda7bab7")
    unexpected_snapshot["id"] = wrong_snap_id
    with pytest.raises(AssertionError, match="is not found"):
        check_snapshot(unexpected_snapshot, swh_storage)

    # 3. snapshot references an inexistent alias
    snapshot0 = Snapshot(
        id=hash_to_bytes("123666f535f882bc7f9a18fb16c9ad27fda7bab7"),
        branches={
            b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
        },
    )
    swh_storage.snapshot_add([snapshot0])

    with pytest.raises(InconsistentAliasBranchError, match="Alias branch HEAD"):
        check_snapshot(snapshot0, swh_storage)

    # 4. snapshot is found in storage, targeted revision does not exist

    rev_not_found = list(swh_storage.revision_missing([REVISION.id]))
    assert len(rev_not_found) == 1

    snapshot1 = Snapshot(
        id=hash_to_bytes("456666f535f882bc7f9a18fb16c9ad27fda7bab7"),
        branches={
            b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
            b"HEAD": SnapshotBranch(
                target=REVISION.id, target_type=TargetType.REVISION,
            ),
        },
    )

    swh_storage.snapshot_add([snapshot1])

    with pytest.raises(InexistentObjectsError, match="Branch/Revision"):
        check_snapshot(snapshot1, swh_storage)

    # 5. snapshot is found in storage, targeted revision exists but the directory the
    # revision targets does not exist

    swh_storage.revision_add([REVISION])

    dir_not_found = list(swh_storage.directory_missing([REVISION.directory]))
    assert len(dir_not_found) == 1

    snapshot2 = Snapshot(
        id=hash_to_bytes("987123f535f882bc7f9a18fb16c9ad27fda7bab7"),
        branches={
            b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
            b"HEAD": SnapshotBranch(
                target=REVISION.id, target_type=TargetType.REVISION,
            ),
        },
    )

    swh_storage.snapshot_add([snapshot2])
    with pytest.raises(InexistentObjectsError, match="Missing directories"):
        check_snapshot(snapshot2, swh_storage)

    assert DIRECTORY.id == REVISION.directory
    swh_storage.directory_add([DIRECTORY])

    # 6. snapshot is found in storage, target revision exists, targeted directory by the
    # revision exist. Content targeted by the directory does not exist.

    assert DIRECTORY.entries[0].target == CONTENT.sha1_git
    not_found = list(swh_storage.content_missing_per_sha1_git([CONTENT.sha1_git]))
    assert len(not_found) == 1

    swh_storage.directory_add([DIRECTORY])

    snapshot3 = Snapshot(
        id=hash_to_bytes("091456f535f882bc7f9a18fb16c9ad27fda7bab7"),
        branches={
            b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
            b"HEAD": SnapshotBranch(
                target=REVISION.id, target_type=TargetType.REVISION,
            ),
        },
    )

    swh_storage.snapshot_add([snapshot3])
    with pytest.raises(InexistentObjectsError, match="Missing content(s)"):
        check_snapshot(snapshot3, swh_storage)

    # 7. snapshot is found in storage, targeted release does not exist

    # release targets the revisions which exists
    assert RELEASE.target == REVISION.id

    snapshot4 = Snapshot(
        id=hash_to_bytes("789666f535f882bc7f9a18fb16c9ad27fda7bab7"),
        branches={
            b"alias": SnapshotBranch(target=b"HEAD", target_type=TargetType.ALIAS,),
            b"HEAD": SnapshotBranch(
                target=REVISION.id, target_type=TargetType.REVISION,
            ),
            b"release/0.1.0": SnapshotBranch(
                target=RELEASE.id, target_type=TargetType.RELEASE,
            ),
        },
    )

    swh_storage.snapshot_add([snapshot4])

    with pytest.raises(InexistentObjectsError, match="Branch/Release"):
        check_snapshot(snapshot4, swh_storage)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task

from swh.loader.package.debian.loader import DebianLoader


@shared_task(name=__name__ + ".LoadDebian")
def load_deb_package(*, url, date, packages):
    """Load Debian package"""
    return DebianLoader(url, date, packages).load()



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
import os

from codecs import BOM_UTF8
from typing import Any, Dict, Generator, Mapping, Sequence, Tuple, Optional

import attr
import chardet

from urllib.parse import quote
from swh.model.model import (
    Person,
    RevisionType,
    Revision,
    TimestampWithTimezone,
    Sha1Git,
)

from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import api_info, release_name


logger = logging.getLogger(__name__)


class NpmLoader(PackageLoader):
    """Load npm origin's artifact releases into swh archive.

    """

    visit_type = "npm"

    def __init__(self, url: str):
        """Constructor

        Args
            str: origin url (e.g. https://www.npmjs.com/package/<package-name>)
        """
        super().__init__(url=url)
        package_name = url.split("https://www.npmjs.com/package/")[1]
        safe_name = quote(package_name, safe="")
        self.provider_url = f"https://replicate.npmjs.com/{safe_name}/"
        self._info: Dict[str, Any] = {}
        self._versions = None

    @property
    def info(self) -> Dict[str, Any]:
        """Return the project metadata information (fetched from npm registry)

        """
        if not self._info:
            self._info = api_info(self.provider_url)
        return self._info

    def get_versions(self) -> Sequence[str]:
        return sorted(list(self.info["versions"].keys()))

    def get_default_version(self) -> str:
        return self.info["dist-tags"].get("latest", "")

    def get_package_info(
        self, version: str
    ) -> Generator[Tuple[str, Mapping[str, Any]], None, None]:
        meta = self.info["versions"][version]
        url = meta["dist"]["tarball"]
        p_info = {
            "url": url,
            "filename": os.path.basename(url),
            "raw": meta,
        }
        yield release_name(version), p_info

    def resolve_revision_from(
        self, known_artifacts: Dict, artifact_metadata: Dict
    ) -> Optional[bytes]:
        return artifact_to_revision_id(known_artifacts, artifact_metadata)

    def build_revision(
        self, a_metadata: Dict, uncompressed_path: str, directory: Sha1Git
    ) -> Optional[Revision]:
        i_metadata = extract_intrinsic_metadata(uncompressed_path)
        if not i_metadata:
            return None
        # from intrinsic metadata
        author = extract_npm_package_author(i_metadata)
        message = i_metadata["version"].encode("ascii")

        # from extrinsic metadata

        # No date available in intrinsic metadata: retrieve it from the API
        # metadata, using the version number that the API claims this package
        # has.
        extrinsic_version = a_metadata["version"]

        if "time" in self.info:
            date = self.info["time"][extrinsic_version]
        elif "mtime" in a_metadata:
            date = a_metadata["mtime"]
        else:
            artifact_name = os.path.basename(a_metadata["dist"]["tarball"])
            raise ValueError(
                "Origin %s: Cannot determine upload time for artifact %s."
                % (self.url, artifact_name)
            )

        date = TimestampWithTimezone.from_iso8601(date)

        # FIXME: this is to remain bug-compatible with earlier versions:
        date = attr.evolve(date, timestamp=attr.evolve(date.timestamp, microseconds=0))

        r = Revision(
            type=RevisionType.TAR,
            message=message,
            author=author,
            date=date,
            committer=author,
            committer_date=date,
            parents=(),
            directory=directory,
            synthetic=True,
            metadata={
                "intrinsic": {"tool": "package.json", "raw": i_metadata,},
                "extrinsic": {
                    "provider": self.provider_url,
                    "when": self.visit_date.isoformat(),
                    "raw": a_metadata,
                },
            },
        )
        return r


def artifact_to_revision_id(
    known_artifacts: Dict, artifact_metadata: Dict
) -> Optional[bytes]:
    """Given metadata artifact, solves the associated revision id.

    The following code allows to deal with 2 metadata formats:

    - old format sample::

        {
            'package_source': {
                'sha1': '05181c12cd8c22035dd31155656826b85745da37',
            }
        }

    - new format sample::

        {
            'original_artifact': [{
                'checksums': {
                    'sha256': '6975816f2c5ad4046acc676ba112f2fff945b01522d63948531f11f11e0892ec', # noqa
                    ...
                },
            }],
            ...
        }

    """
    shasum = artifact_metadata["dist"]["shasum"]
    for rev_id, known_artifact in known_artifacts.items():
        known_original_artifact = known_artifact.get("original_artifact")
        if not known_original_artifact:
            # previous loader-npm version kept original artifact elsewhere
            known_original_artifact = known_artifact.get("package_source")
            if not known_original_artifact:
                continue
            original_hash = known_original_artifact["sha1"]
        else:
            assert isinstance(known_original_artifact, list)
            original_hash = known_original_artifact[0]["checksums"]["sha1"]
        if shasum == original_hash:
            return rev_id
    return None


def extract_npm_package_author(package_json) -> Person:
    """
    Extract package author from a ``package.json`` file content and
    return it in swh format.

    Args:
        package_json (dict): Dict holding the content of parsed
            ``package.json`` file

    Returns:
        Person

    """

    def _author_str(author_data):
        if type(author_data) is dict:
            author_str = ""
            if "name" in author_data:
                author_str += author_data["name"]
            if "email" in author_data:
                author_str += " <%s>" % author_data["email"]
            return author_str
        elif type(author_data) is list:
            return _author_str(author_data[0]) if len(author_data) > 0 else ""
        else:
            return author_data

    for author_key in ("author", "authors"):
        if author_key in package_json:
            author_str = _author_str(package_json[author_key])
            return Person.from_fullname(author_str.encode())

    return Person(fullname=b"", name=None, email=None)


def _lstrip_bom(s, bom=BOM_UTF8):
    if s.startswith(bom):
        return s[len(bom) :]
    else:
        return s


def load_json(json_bytes):
    """
    Try to load JSON from bytes and return a dictionary.

    First try to decode from utf-8. If the decoding failed,
    try to detect the encoding and decode again with replace
    error handling.

    If JSON is malformed, an empty dictionary will be returned.

    Args:
        json_bytes (bytes): binary content of a JSON file

    Returns:
        dict: JSON data loaded in a dictionary
    """
    json_data = {}
    try:
        json_str = _lstrip_bom(json_bytes).decode("utf-8")
    except UnicodeDecodeError:
        encoding = chardet.detect(json_bytes)["encoding"]
        if encoding:
            json_str = json_bytes.decode(encoding, "replace")
    try:
        json_data = json.loads(json_str)
    except json.decoder.JSONDecodeError:
        pass
    return json_data


def extract_intrinsic_metadata(dir_path: str) -> Dict:
    """Given an uncompressed path holding the pkginfo file, returns a
       pkginfo parsed structure as a dict.

       The release artifact contains at their root one folder. For example:
       $ tar tvf zprint-0.0.6.tar.gz
       drwxr-xr-x root/root         0 2018-08-22 11:01 zprint-0.0.6/
       ...

    Args:

        dir_path (str): Path to the uncompressed directory
                        representing a release artifact from npm.

    Returns:
        the pkginfo parsed structure as a dict if any or None if
        none was present.

    """
    # Retrieve the root folder of the archive
    if not os.path.exists(dir_path):
        return {}
    lst = os.listdir(dir_path)
    if len(lst) == 0:
        return {}
    project_dirname = lst[0]
    package_json_path = os.path.join(dir_path, project_dirname, "package.json")
    if not os.path.exists(package_json_path):
        return {}
    with open(package_json_path, "rb") as package_json_file:
        package_json_bytes = package_json_file.read()
        return load_json(package_json_bytes)



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.loader.package.loader import PackageLoader


class FakeStorage:
    def origin_add(self, origins):
        raise ValueError("We refuse to add an origin")

    def origin_visit_get_latest(self, origin):
        return None


class FakeStorage2(FakeStorage):
    def origin_add(self, origins):
        pass

    def origin_visit_add(self, visits):
        raise ValueError("We refuse to add an origin visit")


def test_loader_origin_visit_failure(swh_config):
    """Failure to add origin or origin visit should failed immediately

    """
    loader = PackageLoader("some-url")
    loader.storage = FakeStorage()

    actual_load_status = loader.load()
    assert actual_load_status == {"status": "failed"}

    loader.storage = FakeStorage2()

    actual_load_status2 = loader.load()
    assert actual_load_status2 == {"status": "failed"}



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_archive_loader(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):
    mock_loader = mocker.patch("swh.loader.package.archive.loader.ArchiveLoader.load")
    mock_loader.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.loader.package.archive.tasks.LoadArchive",
        kwargs={"url": "some-url", "artifacts": []},
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task

from swh.loader.package.cran.loader import CRANLoader


@shared_task(name=__name__ + ".LoadCRAN")
def load_cran(url=None, artifacts=[]):
    """Load CRAN's artifacts"""
    return CRANLoader(url, artifacts).load()



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task

from swh.loader.package.deposit.loader import DepositLoader


@shared_task(name=__name__ + ".LoadDeposit")
def load_deposit(*, url, deposit_id):
    """Load Deposit artifacts"""
    return DepositLoader(url, deposit_id).load()



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
import requests
import types

from typing import Any, Dict, Generator, List, Mapping, Optional, Sequence, Tuple, Union

from swh.model.hashutil import hash_to_hex, hash_to_bytes
from swh.model.model import (
    Person,
    Revision,
    RevisionType,
    TimestampWithTimezone,
    Sha1Git,
)
from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import download


logger = logging.getLogger(__name__)


class DepositLoader(PackageLoader):
    """Load pypi origin's artifact releases into swh archive.

    """

    visit_type = "deposit"

    def __init__(self, url: str, deposit_id: str):
        """Constructor

        Args:
            url: Origin url to associate the artifacts/metadata to
            deposit_id: Deposit identity

        """
        super().__init__(url=url)

        config_deposit = self.config["deposit"]
        self.deposit_id = deposit_id
        self.client = ApiClient(url=config_deposit["url"], auth=config_deposit["auth"])
        self.metadata: Dict[str, Any] = {}

    def get_versions(self) -> Sequence[str]:
        # only 1 branch 'HEAD' with no alias since we only have 1 snapshot
        # branch
        return ["HEAD"]

    def get_package_info(
        self, version: str
    ) -> Generator[Tuple[str, Mapping[str, Any]], None, None]:
        p_info = {
            "filename": "archive.zip",
            "raw": self.metadata,
        }
        yield "HEAD", p_info

    def download_package(
        self, p_info: Mapping[str, Any], tmpdir: str
    ) -> List[Tuple[str, Mapping]]:
        """Override to allow use of the dedicated deposit client

        """
        return [self.client.archive_get(self.deposit_id, tmpdir, p_info["filename"])]

    def build_revision(
        self, a_metadata: Dict, uncompressed_path: str, directory: Sha1Git
    ) -> Optional[Revision]:
        depo = a_metadata.pop("deposit")

        # Note:
        # `date` and `committer_date` are always transmitted by the deposit read api
        # which computes itself the values. The loader needs to use those to create the
        # revision.

        # date: codemeta:dateCreated if any, deposit completed_date otherwise
        date = TimestampWithTimezone.from_dict(depo["author_date"])
        # commit_date: codemeta:datePublished if any, deposit completed_date otherwise
        commit_date = TimestampWithTimezone.from_dict(depo["committer_date"])

        client, id, collection = [depo[k] for k in ["client", "id", "collection"]]
        message = f"{client}: Deposit {id} in collection {collection}".encode("utf-8")

        author = parse_author(depo["author"])
        committer = parse_author(depo["committer"])

        return Revision(
            type=RevisionType.TAR,
            message=message,
            author=author,
            date=date,
            committer=committer,
            committer_date=commit_date,
            parents=tuple([hash_to_bytes(p) for p in depo["revision_parents"]]),
            directory=directory,
            synthetic=True,
            metadata={
                "extrinsic": {
                    "provider": self.client.metadata_url(self.deposit_id),
                    "when": self.visit_date.isoformat(),
                    "raw": a_metadata,
                },
            },
        )

    def load(self) -> Dict:
        # First making sure the deposit is known prior to trigger a loading
        try:
            self.metadata = self.client.metadata_get(self.deposit_id)
        except ValueError:
            logger.error(f"Unknown deposit {self.deposit_id}, ignoring")
            return {"status": "failed"}
        # Then usual loading
        r = super().load()
        success = r["status"] != "failed"

        if success:
            # Update archive with metadata information
            origin_metadata = self.metadata["origin_metadata"]
            logger.debug("origin_metadata: %s", origin_metadata)

            provider = origin_metadata["provider"]
            authority = {
                "type": provider["provider_type"],
                "url": provider["provider_url"],
                "metadata": {
                    "name": provider["provider_name"],
                    **(provider["metadata"] or {}),
                },
            }
            self.storage.metadata_authority_add(**authority)

            tool = origin_metadata["tool"]
            fetcher = {
                "name": tool["name"],
                "version": tool["version"],
                "metadata": tool["configuration"],
            }
            self.storage.metadata_fetcher_add(**fetcher)

            metadata = origin_metadata["metadata"]
            format = "sword-v2-atom-codemeta-v2-in-json"
            self.storage.origin_metadata_add(
                self.url,
                self.visit_date,
                {"type": authority["type"], "url": authority["url"]},
                {"name": fetcher["name"], "version": fetcher["version"]},
                format,
                json.dumps(metadata).encode(),
            )

        # Update deposit status
        try:
            if not success:
                self.client.status_update(self.deposit_id, status="failed")
                return r

            snapshot_id = hash_to_bytes(r["snapshot_id"])
            branches = self.storage.snapshot_get(snapshot_id)["branches"]
            logger.debug("branches: %s", branches)
            if not branches:
                return r
            rev_id = branches[b"HEAD"]["target"]

            revisions = self.storage.revision_get([rev_id])
            # FIXME: inconsistency between tests and production code
            if isinstance(revisions, types.GeneratorType):
                revisions = list(revisions)
            revision = revisions[0]

            # Retrieve the revision identifier
            dir_id = revision["directory"]

            # update the deposit's status to success with its
            # revision-id and directory-id
            self.client.status_update(
                self.deposit_id,
                status="done",
                revision_id=hash_to_hex(rev_id),
                directory_id=hash_to_hex(dir_id),
                snapshot_id=r["snapshot_id"],
                origin_url=self.url,
            )
        except Exception:
            logger.exception("Problem when trying to update the deposit's status")
            return {"status": "failed"}
        return r


def parse_author(author) -> Person:
    """See prior fixme

    """
    return Person(
        fullname=author["fullname"].encode("utf-8"),
        name=author["name"].encode("utf-8"),
        email=author["email"].encode("utf-8"),
    )


class ApiClient:
    """Private Deposit Api client

    """

    def __init__(self, url, auth: Optional[Mapping[str, str]]):
        self.base_url = url.rstrip("/")
        self.auth = None if not auth else (auth["username"], auth["password"])

    def do(self, method: str, url: str, *args, **kwargs):
        """Internal method to deal with requests, possibly with basic http
           authentication.

        Args:
            method (str): supported http methods as in get/post/put

        Returns:
            The request's execution output

        """
        method_fn = getattr(requests, method)
        if self.auth:
            kwargs["auth"] = self.auth
        return method_fn(url, *args, **kwargs)

    def archive_get(
        self, deposit_id: Union[int, str], tmpdir: str, filename: str
    ) -> Tuple[str, Dict]:
        """Retrieve deposit's archive artifact locally

        """
        url = f"{self.base_url}/{deposit_id}/raw/"
        return download(url, dest=tmpdir, filename=filename, auth=self.auth)

    def metadata_url(self, deposit_id: Union[int, str]) -> str:
        return f"{self.base_url}/{deposit_id}/meta/"

    def metadata_get(self, deposit_id: Union[int, str]) -> Dict[str, Any]:
        """Retrieve deposit's metadata artifact as json

        """
        url = self.metadata_url(deposit_id)
        r = self.do("get", url)
        if r.ok:
            return r.json()

        msg = f"Problem when retrieving deposit metadata at {url}"
        logger.error(msg)
        raise ValueError(msg)

    def status_update(
        self,
        deposit_id: Union[int, str],
        status: str,
        revision_id: Optional[str] = None,
        directory_id: Optional[str] = None,
        snapshot_id: Optional[str] = None,
        origin_url: Optional[str] = None,
    ):
        """Update deposit's information including status, and persistent
           identifiers result of the loading.

        """
        url = f"{self.base_url}/{deposit_id}/update/"
        payload = {"status": status}
        if revision_id:
            payload["revision_id"] = revision_id
        if directory_id:
            payload["directory_id"] = directory_id
        if snapshot_id:
            payload["snapshot_id"] = snapshot_id
        if origin_url:
            payload["origin_url"] = origin_url

        self.do("put", url, json=payload)



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import json
import logging

import pytest

from json.decoder import JSONDecodeError
from swh.storage.interface import StorageInterface
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from unittest.mock import patch

from swh.model.model import Snapshot, TargetType
from swh.loader.package.archive.loader import ArchiveLoader
from swh.loader.package.nixguix.loader import (
    NixGuixLoader,
    retrieve_sources,
    clean_sources,
)

from swh.loader.package.utils import download
from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.storage.exc import HashCollision

from swh.loader.tests import (
    assert_last_visit_matches,
    get_stats,
    check_snapshot as check_snapshot_full,
)


sources_url = "https://nix-community.github.io/nixpkgs-swh/sources.json"


def check_snapshot(
    snapshot: Union[Dict[str, Any], Snapshot],
    storage: StorageInterface,
    allowed_empty: Iterable[Tuple[TargetType, bytes]] = [
        (TargetType.REVISION, b"evaluation")
    ],
):
    return check_snapshot_full(snapshot, storage, allowed_empty)


def test_retrieve_sources(swh_config, requests_mock_datadir):
    j = retrieve_sources(sources_url)
    assert "sources" in j.keys()
    assert len(j["sources"]) == 2


def test_retrieve_non_existing(swh_config, requests_mock_datadir):
    with pytest.raises(ValueError):
        NixGuixLoader("https://non-existing-url")


def test_retrieve_non_json(swh_config, requests_mock_datadir):
    with pytest.raises(JSONDecodeError):
        NixGuixLoader("https://example.com/file.txt")


def test_clean_sources_invalid_schema(swh_config, requests_mock_datadir):
    sources = {}
    with pytest.raises(ValueError, match="sources structure invalid, missing: .*"):
        clean_sources(sources)


def test_clean_sources_invalid_version(swh_config, requests_mock_datadir):
    for version_ok in [1, "1"]:  # Check those versions are fine
        clean_sources({"version": version_ok, "sources": [], "revision": "my-revision"})

    for version_ko in [0, "0", 2, "2"]:  # Check version != 1 raise an error
        with pytest.raises(
            ValueError, match="sources structure version .* is not supported"
        ):
            clean_sources(
                {"version": version_ko, "sources": [], "revision": "my-revision"}
            )


def test_clean_sources_invalid_sources(swh_config, requests_mock_datadir):
    sources = {
        "version": 1,
        "sources": [
            # Valid source
            {"type": "url", "urls": ["my-url"], "integrity": "my-integrity"},
            # integrity is missing
            {"type": "url", "urls": ["my-url"],},
            # urls is not a list
            {"type": "url", "urls": "my-url", "integrity": "my-integrity"},
            # type is not url
            {"type": "git", "urls": ["my-url"], "integrity": "my-integrity"},
            # missing fields which got double-checked nonetheless...
            {"integrity": "my-integrity"},
        ],
        "revision": "my-revision",
    }
    clean = clean_sources(sources)

    assert len(clean["sources"]) == 1


def check_snapshot_revisions_ok(snapshot, storage):
    """Ensure the snapshot revisions are structurally as expected

    """
    revision_ids = []
    for name, branch in snapshot["branches"].items():
        if name == b"evaluation":
            continue  # skipping that particular branch
        if branch["target_type"] == "revision":
            revision_ids.append(branch["target"])

    revisions = storage.revision_get(revision_ids)
    for rev in revisions:
        metadata = rev["metadata"]
        raw = metadata["extrinsic"]["raw"]
        assert "url" in raw
        assert "integrity" in raw


def test_loader_one_visit(swh_config, requests_mock_datadir):
    loader = NixGuixLoader(sources_url)
    res = loader.load()
    assert res["status"] == "eventful"

    stats = get_stats(loader.storage)
    assert {
        "content": 1,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    # The visit is partial because urls pointing to non tarball file
    # are not handled yet
    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )


def test_uncompress_failure(swh_config, requests_mock_datadir):
    """Non tarball files are currently not supported and the uncompress
    function fails on such kind of files.

    However, even in this case of failure (because of the url
    https://example.com/file.txt), a snapshot and a visit has to be
    created (with a status partial since all files are not archived).

    """
    loader = NixGuixLoader(sources_url)
    loader_status = loader.load()

    urls = [s["urls"][0] for s in loader.sources]
    assert "https://example.com/file.txt" in urls
    assert loader_status["status"] == "eventful"

    # The visit is partial because urls pointing to non tarball files
    # are not handled yet
    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )


def test_loader_incremental(swh_config, requests_mock_datadir):
    """Ensure a second visit do not download artifact already
    downloaded by the previous visit.

    """
    loader = NixGuixLoader(sources_url)
    load_status = loader.load()

    loader.load()
    expected_snapshot_id = "0c5881c74283793ebe9a09a105a9381e41380383"
    assert load_status == {"status": "eventful", "snapshot_id": expected_snapshot_id}

    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )

    expected_branches = {
        b"evaluation": {
            "target": hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            "target_type": "revision",
        },
        b"https://github.com/owner-1/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("488ad4e7b8e2511258725063cf43a2b897c503b4"),
            "target_type": "revision",
        },
    }
    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url == ("https://github.com/owner-1/repository-1/revision-1.tgz")
    ]
    # The artifact
    # 'https://github.com/owner-1/repository-1/revision-1.tgz' is only
    # visited one time
    assert len(urls) == 1


def test_loader_two_visits(swh_config, requests_mock_datadir_visits):
    """To ensure there is only one origin, but two visits, two revisions
    and two snapshots are created.

    The first visit creates a snapshot containing one tarball. The
    second visit creates a snapshot containing the same tarball and
    another tarball.

    """
    loader = NixGuixLoader(sources_url)
    load_status = loader.load()
    expected_snapshot_id = "0c5881c74283793ebe9a09a105a9381e41380383"
    assert load_status == {"status": "eventful", "snapshot_id": expected_snapshot_id}

    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )

    expected_branches = {
        b"evaluation": {
            "target": hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            "target_type": "revision",
        },
        b"https://github.com/owner-1/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("488ad4e7b8e2511258725063cf43a2b897c503b4"),
            "target_type": "revision",
        },
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)

    stats = get_stats(loader.storage)
    assert {
        "content": 1,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": 1,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    loader = NixGuixLoader(sources_url)
    load_status = loader.load()
    expected_snapshot_id = "b0bfa75cbd0cc90aac3b9e95fb0f59c731176d97"
    assert load_status == {"status": "eventful", "snapshot_id": expected_snapshot_id}

    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )

    # This ensures visits are incremental. Indeed, if we request a
    # second time an url, because of the requests_mock_datadir_visits
    # fixture, the file has to end with `_visit1`.
    expected_branches = {
        b"evaluation": {
            "target": hash_to_bytes("602140776b2ce6c9159bcf52ada73a297c063d5e"),
            "target_type": "revision",
        },
        b"https://github.com/owner-1/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("488ad4e7b8e2511258725063cf43a2b897c503b4"),
            "target_type": "revision",
        },
        b"https://github.com/owner-2/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("85e0bad74e33e390aaeb74f139853ae3863ee544"),
            "target_type": "revision",
        },
    }

    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }
    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)

    stats = get_stats(loader.storage)
    assert {
        "content": 2,
        "directory": 5,
        "origin": 1,
        "origin_visit": 2,
        "person": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 2,
    } == stats


def test_resolve_revision_from(swh_config, requests_mock_datadir):
    loader = NixGuixLoader(sources_url)

    known_artifacts = {
        "id1": {"extrinsic": {"raw": {"url": "url1", "integrity": "integrity1"}}},
        "id2": {"extrinsic": {"raw": {"url": "url2", "integrity": "integrity2"}}},
    }

    metadata = {"url": "url1", "integrity": "integrity1"}
    assert loader.resolve_revision_from(known_artifacts, metadata) == "id1"
    metadata = {"url": "url3", "integrity": "integrity3"}
    assert loader.resolve_revision_from(known_artifacts, metadata) == None  # noqa


def test_evaluation_branch(swh_config, requests_mock_datadir):
    loader = NixGuixLoader(sources_url)
    res = loader.load()
    assert res["status"] == "eventful"

    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )

    expected_branches = {
        b"https://github.com/owner-1/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("488ad4e7b8e2511258725063cf43a2b897c503b4"),
            "target_type": "revision",
        },
        b"evaluation": {
            "target": hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            "target_type": "revision",
        },
    }

    expected_snapshot = {
        "id": hash_to_bytes("0c5881c74283793ebe9a09a105a9381e41380383"),
        "branches": expected_branches,
    }

    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)


def test_eoferror(swh_config, requests_mock_datadir):
    """Load a truncated archive which is invalid to make the uncompress
    function raising the exception EOFError. We then check if a
    snapshot is created, meaning this error is well managed.

    """
    sources = (
        "https://nix-community.github.io/nixpkgs-swh/sources-EOFError.json"  # noqa
    )
    loader = NixGuixLoader(sources)
    loader.load()

    expected_branches = {
        b"evaluation": {
            "target": hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            "target_type": "revision",
        },
    }
    expected_snapshot = {
        "id": hash_to_bytes("4257fa2350168c6bfec726a06452ea27a2c0cb33"),
        "branches": expected_branches,
    }

    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)


def fake_download(
    url: str,
    dest: str,
    hashes: Dict = {},
    filename: Optional[str] = None,
    auth: Optional[Tuple[str, str]] = None,
) -> Tuple[str, Dict]:
    """Fake download which raises HashCollision (for the sake of test simpliciy,
    let's accept that makes sense)

    For tests purpose only.

    """
    if url == "https://example.com/file.txt":
        # instead of failing because it's a file not dealt with by the nix guix
        # loader, make it raise a hash collision
        raise HashCollision("sha1", "f92d74e3874587aaf443d1db961d4e26dde13e9c", [])
    return download(url, dest, hashes, filename, auth)


def test_raise_exception(swh_config, requests_mock_datadir, mocker):
    mock_download = mocker.patch("swh.loader.package.loader.download")
    mock_download.side_effect = fake_download

    loader = NixGuixLoader(sources_url)
    res = loader.load()

    expected_snapshot_id = "0c5881c74283793ebe9a09a105a9381e41380383"
    assert res == {
        "status": "eventful",
        "snapshot_id": expected_snapshot_id,
    }

    expected_branches = {
        b"https://github.com/owner-1/repository-1/revision-1.tgz": {
            "target": hash_to_bytes("488ad4e7b8e2511258725063cf43a2b897c503b4"),
            "target_type": "revision",
        },
        b"evaluation": {
            "target": hash_to_bytes("cc4e04c26672dd74e5fd0fecb78b435fb55368f7"),
            "target_type": "revision",
        },
    }
    expected_snapshot = {
        "id": hash_to_bytes(expected_snapshot_id),
        "branches": expected_branches,
    }

    snapshot = check_snapshot(expected_snapshot, storage=loader.storage)
    check_snapshot_revisions_ok(snapshot, loader.storage)

    assert len(mock_download.mock_calls) == 2

    # The visit is partial because some artifact downloads failed
    assert_last_visit_matches(
        loader.storage, sources_url, status="partial", type="nixguix"
    )


def test_load_nixguix_one_common_artifact_from_other_loader(
    swh_config, datadir, requests_mock_datadir_visits, caplog
):
    """Misformatted revision should be caught and logged, then loading continues

    """
    caplog.set_level(logging.ERROR, "swh.loader.package.nixguix.loader")

    # 1. first ingest with for example the archive loader
    gnu_url = "https://ftp.gnu.org/gnu/8sync/"
    release = "0.1.0"
    artifact_url = f"https://ftp.gnu.org/gnu/8sync/8sync-{release}.tar.gz"
    gnu_artifacts = [
        {
            "time": 944729610,
            "url": artifact_url,
            "length": 221837,
            "filename": f"8sync-{release}.tar.gz",
            "version": release,
        }
    ]
    archive_loader = ArchiveLoader(url=gnu_url, artifacts=gnu_artifacts)
    actual_load_status = archive_loader.load()
    expected_snapshot_id = "c419397fd912039825ebdbea378bc6283f006bf5"
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] == expected_snapshot_id  # noqa

    assert_last_visit_matches(
        archive_loader.storage, gnu_url, status="full", type="tar"
    )

    gnu_snapshot = archive_loader.storage.snapshot_get(
        hash_to_bytes(expected_snapshot_id)
    )

    first_revision = gnu_snapshot["branches"][f"releases/{release}".encode("utf-8")]

    # 2. Then ingest with the nixguix loader which lists the same artifact within its
    # sources.json

    # ensure test setup is ok
    data_sources = os.path.join(
        datadir, "https_nix-community.github.io", "nixpkgs-swh_sources_special.json"
    )
    all_sources = json.loads(open(data_sources).read())
    found = False
    for source in all_sources["sources"]:
        if source["urls"][0] == artifact_url:
            found = True
            assert (
                found is True
            ), f"test setup error: {artifact_url} must be in {data_sources}"

    # first visit with a snapshot, ok
    sources_url = "https://nix-community.github.io/nixpkgs-swh/sources_special.json"
    loader = NixGuixLoader(sources_url)
    actual_load_status2 = loader.load()
    assert actual_load_status2["status"] == "eventful"

    assert_last_visit_matches(
        loader.storage, sources_url, status="full", type="nixguix"
    )

    snapshot_id = actual_load_status2["snapshot_id"]
    snapshot = loader.storage.snapshot_get(hash_to_bytes(snapshot_id))
    snapshot.pop("next_branch")  # snapshot_get endpoint detail to drop

    # simulate a snapshot already seen with a revision with the wrong metadata structure
    # This revision should be skipped, thus making the artifact being ingested again.
    with patch(
        "swh.loader.package.loader.PackageLoader.last_snapshot"
    ) as last_snapshot:
        # mutate the snapshot to target a revision with the wrong metadata structure
        # snapshot["branches"][artifact_url.encode("utf-8")] = first_revision
        old_revision = next(loader.storage.revision_get([first_revision["target"]]))
        # assert that revision is not in the right format
        assert old_revision["metadata"]["extrinsic"]["raw"].get("integrity", {}) == {}

        # mutate snapshot to create a clash
        snapshot["branches"][artifact_url.encode("utf-8")] = {
            "target_type": "revision",
            "target": hash_to_bytes(old_revision["id"]),
        }

        # modify snapshot to actually change revision metadata structure so we simulate
        # a revision written by somebody else (structure different)
        last_snapshot.return_value = Snapshot.from_dict(snapshot)

        loader = NixGuixLoader(sources_url)
        actual_load_status3 = loader.load()
        assert last_snapshot.called
        assert actual_load_status3["status"] == "eventful"

        assert_last_visit_matches(
            loader.storage, sources_url, status="full", type="nixguix"
        )

        new_snapshot_id = "32ff641e510aceefc3a6d0dcbf208b2854d2e965"
        assert actual_load_status3["snapshot_id"] == new_snapshot_id

        last_snapshot = loader.storage.snapshot_get(hash_to_bytes(new_snapshot_id))
        new_revision_branch = last_snapshot["branches"][artifact_url.encode("utf-8")]
        assert new_revision_branch["target_type"] == "revision"

        new_revision = next(
            loader.storage.revision_get([new_revision_branch["target"]])
        )

        # the new revision has the correct structure,  so it got ingested alright by the
        # new run
        assert new_revision["metadata"]["extrinsic"]["raw"]["integrity"] is not None

        nb_detections = 0
        actual_detection: Dict
        for record in caplog.records:
            logtext = record.getMessage()
            if "Unexpected metadata revision structure detected:" in logtext:
                nb_detections += 1
                actual_detection = record.args["context"]

        assert actual_detection
        # as many calls as there are sources listed in the sources.json
        assert nb_detections == len(all_sources["sources"])

        assert actual_detection == {
            "revision": hash_to_hex(old_revision["id"]),
            "reason": "'integrity'",
            "known_artifact": old_revision["metadata"],
        }



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model.hashutil import hash_to_bytes

from swh.loader.package.archive.loader import ArchiveLoader
from swh.loader.package.tests.common import check_metadata_paths
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
)


URL = "https://ftp.gnu.org/gnu/8sync/"
GNU_ARTIFACTS = [
    {
        "time": 944729610,
        "url": "https://ftp.gnu.org/gnu/8sync/8sync-0.1.0.tar.gz",
        "length": 221837,
        "filename": "8sync-0.1.0.tar.gz",
        "version": "0.1.0",
    }
]

_expected_new_contents_first_visit = [
    "e9258d81faf5881a2f96a77ba609396f82cb97ad",
    "1170cf105b04b7e2822a0e09d2acf71da7b9a130",
    "fbd27c3f41f2668624ffc80b7ba5db9b92ff27ac",
    "0057bec9b5422aff9256af240b177ac0e3ac2608",
    "2b8d0d0b43a1078fc708930c8ddc2956a86c566e",
    "27de3b3bc6545d2a797aeeb4657c0e215a0c2e55",
    "2e6db43f5cd764e677f416ff0d0c78c7a82ef19b",
    "ae9be03bd2a06ed8f4f118d3fe76330bb1d77f62",
    "edeb33282b2bffa0e608e9d2fd960fd08093c0ea",
    "d64e64d4c73679323f8d4cde2643331ba6c20af9",
    "7a756602914be889c0a2d3952c710144b3e64cb0",
    "84fb589b554fcb7f32b806951dcf19518d67b08f",
    "8624bcdae55baeef00cd11d5dfcfa60f68710a02",
    "e08441aeab02704cfbd435d6445f7c072f8f524e",
    "f67935bc3a83a67259cda4b2d43373bd56703844",
    "809788434b433eb2e3cfabd5d591c9a659d5e3d8",
    "7d7c6c8c5ebaeff879f61f37083a3854184f6c41",
    "b99fec102eb24bffd53ab61fc30d59e810f116a2",
    "7d149b28eaa228b3871c91f0d5a95a2fa7cb0c68",
    "f0c97052e567948adf03e641301e9983c478ccff",
    "7fb724242e2b62b85ca64190c31dcae5303e19b3",
    "4f9709e64a9134fe8aefb36fd827b84d8b617ab5",
    "7350628ccf194c2c3afba4ac588c33e3f3ac778d",
    "0bb892d9391aa706dc2c3b1906567df43cbe06a2",
    "49d4c0ce1a16601f1e265d446b6c5ea6b512f27c",
    "6b5cc594ac466351450f7f64a0b79fdaf4435ad3",
    "3046e5d1f70297e2a507b98224b6222c9688d610",
    "1572607d456d7f633bc6065a2b3048496d679a31",
]

_expected_new_directories_first_visit = [
    "daabc65ec75d487b1335ffc101c0ac11c803f8fc",
    "263be23b4a8101d3ad0d9831319a3e0f2b065f36",
    "7f6e63ba6eb3e2236f65892cd822041f1a01dd5c",
    "4db0a3ecbc976083e2dac01a62f93729698429a3",
    "dfef1c80e1098dd5deda664bb44a9ab1f738af13",
    "eca971d346ea54d95a6e19d5051f900237fafdaa",
    "3aebc29ed1fccc4a6f2f2010fb8e57882406b528",
]

_expected_new_revisions_first_visit = {
    "44183488c0774ce3c957fa19ba695cf18a4a42b3": (
        "3aebc29ed1fccc4a6f2f2010fb8e57882406b528"
    )
}

_expected_branches_first_visit = {
    b"HEAD": {"target_type": "alias", "target": b"releases/0.1.0",},
    b"releases/0.1.0": {
        "target_type": "revision",
        "target": hash_to_bytes("44183488c0774ce3c957fa19ba695cf18a4a42b3"),
    },
}

# hash is different then before as we changed the snapshot
# gnu used to use `release/` (singular) instead of plural
_expected_new_snapshot_first_visit_id = hash_to_bytes(
    "c419397fd912039825ebdbea378bc6283f006bf5"
)


def visit_with_no_artifact_found(swh_config, requests_mock_datadir):
    url = URL
    unknown_artifact_url = "https://ftp.g.o/unknown/8sync-0.1.0.tar.gz"
    loader = ArchiveLoader(
        url,
        artifacts=[
            {
                "time": 944729610,
                "url": unknown_artifact_url,  # unknown artifact
                "length": 221837,
                "filename": "8sync-0.1.0.tar.gz",
                "version": "0.1.0",
            }
        ],
    )

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "uneventful"
    assert actual_load_status["snapshot_id"] is not None
    stats = get_stats(loader.storage)

    assert {
        "content": 0,
        "directory": 0,
        "origin": 1,
        "origin_visit": 1,
        "person": 0,
        "release": 0,
        "revision": 0,
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    assert_last_visit_matches(loader.storage, url, status="partial", type="tar")


def test_check_revision_metadata_structure(swh_config, requests_mock_datadir):
    loader = ArchiveLoader(url=URL, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None

    assert_last_visit_matches(loader.storage, URL, status="full", type="tar")

    expected_revision_id = hash_to_bytes("44183488c0774ce3c957fa19ba695cf18a4a42b3")
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(
        revision["metadata"],
        paths=[
            ("intrinsic", dict),
            ("extrinsic.provider", str),
            ("extrinsic.when", str),
            ("extrinsic.raw", dict),
            ("original_artifact", list),
        ],
    )

    for original_artifact in revision["metadata"]["original_artifact"]:
        check_metadata_paths(
            original_artifact,
            paths=[("filename", str), ("length", int), ("checksums", dict),],
        )


def test_visit_with_release_artifact_no_prior_visit(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    loader = ArchiveLoader(url=URL, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert (
        hash_to_bytes(actual_load_status["snapshot_id"])
        == _expected_new_snapshot_first_visit_id
    )  # noqa

    assert_last_visit_matches(loader.storage, URL, status="full", type="tar")

    stats = get_stats(loader.storage)
    assert {
        "content": len(_expected_new_contents_first_visit),
        "directory": len(_expected_new_directories_first_visit),
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": len(_expected_new_revisions_first_visit),
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    expected_contents = map(hash_to_bytes, _expected_new_contents_first_visit)
    assert list(loader.storage.content_missing_per_sha1(expected_contents)) == []

    expected_dirs = map(hash_to_bytes, _expected_new_directories_first_visit)
    assert list(loader.storage.directory_missing(expected_dirs)) == []

    expected_revs = map(hash_to_bytes, _expected_new_revisions_first_visit)
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_snapshot = {
        "id": _expected_new_snapshot_first_visit_id,
        "branches": _expected_branches_first_visit,
    }

    check_snapshot(expected_snapshot, loader.storage)


def test_2_visits_without_change(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    url = URL
    loader = ArchiveLoader(url, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None

    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    actual_load_status2 = loader.load()
    assert actual_load_status2["status"] == "uneventful"
    assert actual_load_status2["snapshot_id"] is not None
    assert actual_load_status["snapshot_id"] == actual_load_status2["snapshot_id"]

    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url.startswith("https://ftp.gnu.org")
    ]
    assert len(urls) == 1


def test_2_visits_with_new_artifact(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    url = URL
    artifact1 = GNU_ARTIFACTS[0]
    loader = ArchiveLoader(url, [artifact1])

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None

    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    stats = get_stats(loader.storage)
    assert {
        "content": len(_expected_new_contents_first_visit),
        "directory": len(_expected_new_directories_first_visit),
        "origin": 1,
        "origin_visit": 1,
        "person": 1,
        "release": 0,
        "revision": len(_expected_new_revisions_first_visit),
        "skipped_content": 0,
        "snapshot": 1,
    } == stats

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url.startswith("https://ftp.gnu.org")
    ]
    assert len(urls) == 1

    artifact2 = {
        "time": 1480991830,
        "url": "https://ftp.gnu.org/gnu/8sync/8sync-0.2.0.tar.gz",
        "length": 238466,
        "filename": "8sync-0.2.0.tar.gz",
        "version": "0.2.0",
    }

    loader2 = ArchiveLoader(url, [artifact1, artifact2])
    # implementation detail: share the storage in between visits
    loader2.storage = loader.storage
    stats2 = get_stats(loader2.storage)
    assert stats == stats2  # ensure we share the storage

    actual_load_status2 = loader2.load()
    assert actual_load_status2["status"] == "eventful"
    assert actual_load_status2["snapshot_id"] is not None

    stats2 = get_stats(loader.storage)
    assert {
        "content": len(_expected_new_contents_first_visit) + 14,
        "directory": len(_expected_new_directories_first_visit) + 8,
        "origin": 1,
        "origin_visit": 1 + 1,
        "person": 1,
        "release": 0,
        "revision": len(_expected_new_revisions_first_visit) + 1,
        "skipped_content": 0,
        "snapshot": 1 + 1,
    } == stats2

    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url.startswith("https://ftp.gnu.org")
    ]
    # 1 artifact (2nd time no modification) + 1 new artifact
    assert len(urls) == 2


def test_2_visits_without_change_not_gnu(swh_config, requests_mock_datadir):
    """Load a project archive (not gnu) ends up with 1 snapshot

    """
    url = "https://something.else.org/8sync/"
    artifacts = [  # this is not a gnu artifact
        {
            "time": "1999-12-09T09:53:30+00:00",  # it's also not a timestamp
            "sha256": "d5d1051e59b2be6f065a9fc6aedd3a391e44d0274b78b9bb4e2b57a09134dbe4",  # noqa
            # keep a gnu artifact reference to avoid adding other test files
            "url": "https://ftp.gnu.org/gnu/8sync/8sync-0.2.0.tar.gz",
            "length": 238466,
            "filename": "8sync-0.2.0.tar.gz",
            "version": "0.2.0",
        }
    ]

    # Here the loader defines the id_keys to use for existence in the snapshot
    # It's not the default archive loader which
    loader = ArchiveLoader(
        url, artifacts=artifacts, identity_artifact_keys=["sha256", "length", "url"]
    )

    actual_load_status = loader.load()
    assert actual_load_status["status"] == "eventful"
    assert actual_load_status["snapshot_id"] is not None
    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    actual_load_status2 = loader.load()
    assert actual_load_status2["status"] == "uneventful"
    assert actual_load_status2["snapshot_id"] == actual_load_status["snapshot_id"]
    assert_last_visit_matches(loader.storage, url, status="full", type="tar")

    urls = [
        m.url
        for m in requests_mock_datadir.request_history
        if m.url.startswith("https://ftp.gnu.org")
    ]
    assert len(urls) == 1



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task

from swh.loader.package.archive.loader import ArchiveLoader


@shared_task(name=__name__ + ".LoadArchive")
def load_archive_files(*, url=None, artifacts=None, identity_artifact_keys=None):
    """Load archive's artifacts (e.g gnu, etc...)"""
    return ArchiveLoader(
        url, artifacts, identity_artifact_keys=identity_artifact_keys
    ).load()



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

import pytest
import yaml

from typing import Any, Dict


@pytest.fixture
def swh_loader_config(swh_storage_postgresql) -> Dict[str, Any]:
    return {
        "storage": {
            "cls": "pipeline",
            "steps": [
                {"cls": "retry"},
                {"cls": "filter"},
                {"cls": "buffer"},
                {
                    "cls": "local",
                    "db": swh_storage_postgresql.dsn,
                    "objstorage": {"cls": "memory", "args": {}},
                },
            ],
        },
        "deposit": {
            "url": "https://deposit.softwareheritage.org/1/private",
            "auth": {"username": "user", "password": "pass",},
        },
    }


@pytest.fixture
def swh_config(swh_loader_config, monkeypatch, tmp_path):
    conffile = os.path.join(str(tmp_path), "loader.yml")
    with open(conffile, "w") as f:
        f.write(yaml.dump(swh_loader_config))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conffile)
    return conffile


@pytest.fixture(autouse=True, scope="session")
def swh_proxy():
    """Automatically inject this fixture in all tests to ensure no outside
       connection takes place.

    """
    os.environ["http_proxy"] = "http://localhost:999"
    os.environ["https_proxy"] = "http://localhost:999"



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os import path

import logging

from typing import Dict, List, Tuple

logger = logging.getLogger(__file__)


DATADIR = path.join(path.abspath(path.dirname(__file__)), "resources")


def check_metadata(metadata: Dict, key_path: str, raw_type: str):
    """Given a metadata dict, ensure the associated key_path value is of type
       raw_type.

    Args:
        metadata: Dict to check
        key_path: Path to check
        raw_type: Type to check the path with

    Raises:
        Assertion error in case of mismatch

    """
    data = metadata
    keys = key_path.split(".")
    for k in keys:
        try:
            data = data[k]
        except (TypeError, KeyError) as e:
            # KeyError: because path too long
            # TypeError: data is not a dict
            raise AssertionError(e)
    assert isinstance(data, raw_type)  # type: ignore


def check_metadata_paths(metadata: Dict, paths: List[Tuple[str, str]]):
    """Given a metadata dict, ensure the keys are of expected types

    Args:
        metadata: Dict to check
        key_path: Path to check
        raw_type: Type to check the path with

    Raises:
        Assertion error in case of mismatch

    """
    for key_path, raw_type in paths:
        check_metadata(metadata, key_path, raw_type)
